"""Engine parser."""
import xml.etree.ElementTree as etree

import pandas as pd
import regex as re
from loguru import logger

from pyiohat.parsers.ident_base_parser import IdentBaseParser


class MSGFPlus_2021_03_22_Parser(IdentBaseParser):
    """File parser for MSGF+."""

    def __init__(self, *args, **kwargs):
        """Initialize parser.

        Reads in data file and provides mappings.
        """
        super().__init__(*args, **kwargs)
        self.version = None
        self.peptide_lookup = None
        self.spec_records = None
        self.style = "msgfplus_style_1"
        self.mapping_dict = {
            v: k
            for k, v in self.param_mapper.get_default_params(style=self.style)[
                "header_translations"
            ]["translated_value"].items()
        }

    @classmethod
    def check_parser_compatibility(cls, file):
        """Assert compatibility between file and parser.

        Args:
            file (path object): path to input file

        Returns:
            bool: True if parser and file are compatible
        """
        is_mzid = file.name.endswith(".mzid")

        with open(file) as f:
            try:
                head = "".join([next(f) for _ in range(20)])
            except StopIteration:
                head = ""
        contains_engine = "MS-GF+" in head
        return is_mzid and contains_engine

    def get_version(self):
        """Retrieve version from xml.

        Returns:
            version (str): file version
        """
        version = ""
        for event, entry in etree.iterparse(self.input_file):
            entry_tag = entry.tag

            if entry_tag.endswith("cvList"):
                if entry_tag != "{http://psidev.info/psi/pi/mzIdentML/1.1}cvList":
                    logger.warning(
                        f"{entry_tag}: Wrong mzIdentML version - Parser made for version 1.1.0!"
                    )
            elif entry_tag.endswith("AnalysisSoftware"):
                version = "msgfplus_" + "_".join(
                    re.findall("[0-9]+", entry.attrib["version"])
                )
                break
            entry.clear()
        return version

    def get_peptide_lookup(self):
        """Retrieve peptide ids with their corresponding sequences and modifications from xml.

        Returns:
            peptide_lookup (dict): peptide id with corresponding modifications and sequence
        """
        peptide_lookup = {}

        cv_param_modifications = ""
        peptide_information = False

        for event, entry in etree.iterparse(self.input_file):
            entry_tag = entry.tag

            if entry_tag.endswith("PeptideSequence"):
                peptide_information = True
            if peptide_information is True:
                if entry_tag.endswith("PeptideSequence"):
                    sequence = {"sequence": entry.text}
                elif entry_tag.endswith("cvParam"):
                    if entry.attrib["name"] == "unknown modification":
                        cv_param_modifications += entry.attrib["value"] + ":"
                    else:
                        cv_param_modifications += entry.attrib["name"] + ":"
                elif entry_tag.endswith("Modification"):
                    cv_param_modifications += entry.attrib["location"] + ";"
                elif entry_tag.endswith("Peptide"):
                    peptide_lookup[entry.attrib["id"]] = sequence
                    peptide_lookup[entry.attrib["id"]][
                        "modifications"
                    ] = cv_param_modifications.rstrip(";")
                    cv_param_modifications = ""
                elif entry_tag.endswith("PeptideEvidence"):
                    break
            entry.clear()
        return peptide_lookup

    def get_spec_records(self):
        """Retrieve specs from file.

        Returns:
            spec_records (list): information on PSMs
        """
        spec_records = []

        spec_results = {}
        spec_ident_items = []
        spec_information = False

        for event, entry in etree.iterparse(self.input_file):
            entry_tag = entry.tag

            if entry_tag.endswith("FragmentationTable"):
                spec_information = True
            elif spec_information is True:
                if entry_tag.endswith("cvParam") or entry_tag.endswith("userParam"):
                    if entry.attrib["name"] in self.mapping_dict:
                        _key = self.mapping_dict[entry.attrib["name"]]
                        spec_results[_key] = entry.attrib["value"]
                elif entry_tag.endswith("SpectrumIdentificationItem"):
                    for attribute in list(entry.attrib):
                        if attribute in self.mapping_dict.keys():
                            if attribute == "peptide_ref":
                                psm_id = entry.attrib[attribute]
                                spec_results["sequence"] = self.peptide_lookup[psm_id][
                                    "sequence"
                                ]
                                spec_results["modifications"] = self.peptide_lookup[
                                    psm_id
                                ]["modifications"]
                            else:
                                _key = self.mapping_dict[attribute]
                                spec_results[_key] = entry.attrib[attribute]
                    # multiple SpectrumIdentificationItems possible, therefore create a list and reset spec_results
                    spec_ident_items.append(spec_results)
                    spec_results = {}
                elif entry_tag.endswith("SpectrumIdentificationResult"):
                    for spec_item in spec_ident_items:
                        spec_item.update(spec_results)
                        spec_records.append(spec_item)
                    spec_results = {}
                    spec_ident_items = []
                elif entry_tag.endswith("SpectrumIdentificationList"):
                    break
            entry.clear()
        return spec_records

    def unify(self):
        """
        Primary method to read and unify engine output.

        Returns:
            self.df (pd.DataFrame): unified dataframe
        """
        self.version = self.get_version()
        self.peptide_lookup = self.get_peptide_lookup()
        self.spec_records = self.get_spec_records()
        self.df = pd.DataFrame(self.spec_records)
        self.df["search_engine"] = self.version
        self.process_unify_style()

        return self.df



#!/usr/bin/env python
import pandas as pd
import pytest

from pyiohat.parsers.ident.xtandem_alanine import (
    XTandemAlanine_Parser,
)


def test_engine_parsers_xtandem_init():
    input_file = (
        pytest._test_path / "data" / "test_Creinhardtii_QE_pH11_xtandem_alanine.xml"
    )

    parser = XTandemAlanine_Parser(
        input_file,
        params={
            "cpus": 2,
            "enzyme": "(?<=[KR])(?![P])",
            "terminal_cleavage_site_integrity": "any",
            "validation_score_field": {"xtandem_alanine": "x!tandem:hyperscore"},
            "bigger_scores_better": {"xtandem_alanine": True},
            "modifications": [
                {
                    "aa": "M",
                    "type": "opt",
                    "position": "any",
                    "name": "Oxidation",
                },
                {
                    "aa": "C",
                    "type": "fix",
                    "position": "any",
                    "name": "Carbamidomethyl",
                },
                {
                    "aa": "*",
                    "type": "opt",
                    "position": "Prot-N-term",
                    "name": "Acetyl",
                },
            ],
        },
    )


def test_engine_parsers_xtandem_file_matches_xtandem_parser():
    input_file = (
        pytest._test_path / "data" / "test_Creinhardtii_QE_pH11_xtandem_alanine.xml"
    )

    assert XTandemAlanine_Parser.check_parser_compatibility(input_file) is True


def test_engine_parsers_msfragger_file_not_matches_xtandem_parser():
    input_file = (
        pytest._test_path / "data" / "test_Creinhardtii_QE_pH11_msfragger_3.tsv"
    )

    assert XTandemAlanine_Parser.check_parser_compatibility(input_file) is False


def test_engine_parsers_xtandem_check_dataframe_integrity():
    input_file = (
        pytest._test_path / "data" / "test_Creinhardtii_QE_pH11_xtandem_alanine.xml"
    )
    rt_lookup_path = pytest._test_path / "data" / "_ursgal_lookup.csv"
    db_path = pytest._test_path / "data" / "test_Creinhardtii_target_decoy.fasta"

    parser = XTandemAlanine_Parser(
        input_file,
        params={
            "cpus": 2,
            "rt_pickle_name": rt_lookup_path,
            "database": db_path,
            "enzyme": "(?<=[KR])(?![P])",
            "terminal_cleavage_site_integrity": "any",
            "validation_score_field": {"xtandem_alanine": "x!tandem:hyperscore"},
            "bigger_scores_better": {"xtandem_alanine": True},
            "modifications": [
                {
                    "aa": "M",
                    "type": "opt",
                    "position": "any",
                    "name": "Oxidation",
                },
                {
                    "aa": "C",
                    "type": "fix",
                    "position": "any",
                    "name": "Carbamidomethyl",
                },
                {
                    "aa": "*",
                    "type": "opt",
                    "position": "Prot-N-term",
                    "name": "Acetyl",
                },
            ],
        },
    )
    df = parser.unify()
    assert len(df) == 79
    assert (df["raw_data_location"] == "path/for/glory.mzML").all()
    assert pytest.approx(df["ucalc_mz"].mean()) == 796.4324
    assert pytest.approx(df["exp_mz"].mean()) == 796.71967

    assert df["modifications"].str.contains("Acetyl:0").sum() == 1
    assert df["modifications"].str.contains("Oxidation:").sum() == 23
    assert (
        df["modifications"].str.count("Carbamidomethyl:")
        == df["sequence"].str.count("C")
    ).all()
    assert df["modifications"].str.count(":").sum() == 50


def test_engine_parsers_xtandem_nterminal_mod():
    input_file = (
        pytest._test_path / "data" / "test_Creinhardtii_QE_pH11_xtandem_alanine.xml"
    )
    rt_lookup_path = pytest._test_path / "data" / "_ursgal_lookup.csv"
    db_path = pytest._test_path / "data" / "test_Creinhardtii_target_decoy.fasta"

    parser = XTandemAlanine_Parser(
        input_file,
        params={
            "cpus": 2,
            "rt_pickle_name": rt_lookup_path,
            "database": db_path,
            "enzyme": "(?<=[KR])(?![P])",
            "terminal_cleavage_site_integrity": "any",
            "validation_score_field": {"xtandem_alanine": "x!tandem:hyperscore"},
            "bigger_scores_better": {"xtandem_alanine": True},
            "modifications": [
                {
                    "aa": "M",
                    "type": "opt",
                    "position": "any",
                    "name": "Oxidation",
                },
                {
                    "aa": "C",
                    "type": "fix",
                    "position": "any",
                    "name": "Carbamidomethyl",
                },
                {
                    "aa": "*",
                    "type": "opt",
                    "position": "Prot-N-term",
                    "name": "Acetyl",
                },
            ],
            "raw_file_location": "test_Creinhardtii_QE_pH11.mzML",
            "15N": False,
        },
    )
    df = parser.unify()
    relevant_row = df[df["sequence"] == "WGLVSSELQTSEAETPGLK"]
    assert relevant_row["modifications"].tolist() == ["Acetyl:0"]


def test_engine_parsers_xtandem_multiple_psms():
    input_file = pytest._test_path / "data" / "multiple_psms_xtandem.xml"
    rt_lookup_path = pytest._test_path / "data" / "_ursgal_lookup.csv"
    db_path = pytest._test_path / "data" / "human_ecoli_test_target_decoy.fasta"

    parser = XTandemAlanine_Parser(
        input_file,
        params={
            "cpus": 2,
            "rt_pickle_name": rt_lookup_path,
            "database": db_path,
            "enzyme": "(?<=[KR])(?![P])",
            "terminal_cleavage_site_integrity": "any",
            "validation_score_field": {"xtandem_alanine": "x!tandem:hyperscore"},
            "bigger_scores_better": {"xtandem_alanine": True},
            "modifications": [
                {
                    "aa": "M",
                    "type": "opt",
                    "position": "any",
                    "name": "Oxidation",
                },
                {
                    "aa": "C",
                    "type": "fix",
                    "position": "any",
                    "name": "Carbamidomethyl",
                },
                {
                    "aa": "*",
                    "type": "opt",
                    "position": "Prot-N-term",
                    "name": "Acetyl",
                },
            ],
            "raw_file_location": "test_Creinhardtii_QE_pH11.mzML",
            "15N": False,
        },
    )

    # Test file with:
    #    - one sequence in first group
    #    - 3 sequences in second group

    df = parser.unify()
    assert len(df) == 4

    assert set(df["sequence"]) == {
        "ITIPITLRMLIAK",
        "SMMNGGSSPESDVGTDNK",
        "SMMNGGSSPESDVGTDNK",
        "SMMNGGSSPESDVGTDNK",
    }
    assert set(df["spectrum_id"]) == {12833, 14525}
    assert set(df["modifications"]) == {
        "Acetyl:0",
        "",
        "Oxidation:2",
        "Oxidation:3",
    }


def test_engine_parsers_xtandem_map_mod_names():
    input_file = (
        pytest._test_path / "data" / "test_Creinhardtii_QE_pH11_xtandem_alanine.xml"
    )

    parser = XTandemAlanine_Parser(
        input_file,
        params={
            "cpus": 2,
            "enzyme": "(?<=[KR])(?![P])",
            "terminal_cleavage_site_integrity": "any",
            "validation_score_field": {"xtandem_alanine": "x!tandem:hyperscore"},
            "bigger_scores_better": {"xtandem_alanine": True},
            "modifications": [
                {
                    "aa": "M",
                    "type": "opt",
                    "position": "any",
                    "name": "Oxidation",
                },
                {
                    "aa": "C",
                    "type": "fix",
                    "position": "any",
                    "name": "Carbamidomethyl",
                },
                {
                    "aa": "*",
                    "type": "opt",
                    "position": "Prot-N-term",
                    "name": "Acetyl",
                },
            ],
        },
    )
    test_df = pd.DataFrame({"modifications": [["57.021464:0"]], "sequence": ["CERK"]})
    assert parser.map_mod_names(test_df)["modifications"][0] == "Carbamidomethyl:1"


def test_engine_parsers_xtandem_map_mod_names_nterm():
    input_file = (
        pytest._test_path / "data" / "test_Creinhardtii_QE_pH11_xtandem_alanine.xml"
    )

    parser = XTandemAlanine_Parser(
        input_file,
        params={
            "cpus": 2,
            "enzyme": "(?<=[KR])(?![P])",
            "terminal_cleavage_site_integrity": "any",
            "validation_score_field": {"xtandem_alanine": "x!tandem:hyperscore"},
            "bigger_scores_better": {"xtandem_alanine": True},
            "modifications": [
                {
                    "aa": "M",
                    "type": "opt",
                    "position": "any",
                    "name": "Oxidation",
                },
                {
                    "aa": "C",
                    "type": "fix",
                    "position": "any",
                    "name": "Carbamidomethyl",
                },
                {
                    "aa": "*",
                    "type": "opt",
                    "position": "Prot-N-term",
                    "name": "Acetyl",
                },
            ],
        },
    )

    row = pd.DataFrame(
        {"modifications": [["57.021464:0", "42.010565:0"]], "sequence": ["CERK"]}
    )
    assert set(parser.map_mod_names(row)["modifications"][0].split(";")) == {
        "Carbamidomethyl:1",
        "Acetyl:0",
    }


def test_engine_parsers_xtandem_get_spec_records():
    input_file = (
        pytest._test_path / "data" / "test_Creinhardtii_QE_pH11_xtandem_alanine.xml"
    )
    parser = XTandemAlanine_Parser(input_file=input_file, params=None)
    spec_records, search_engine = parser.get_spec_records()

    assert len(spec_records) == 79
    assert spec_records[0] == {
        "x!tandem:delta": "0.0057",
        "x!tandem:hyperscore": "14.2",
        "x!tandem:nextscore": "8.0",
        "x!tandem:y_score": "9.9",
        "x!tandem:y_ions": "5",
        "x!tandem:b_score": "0.0",
        "x!tandem:b_ions": "0",
        "sequence": "DDVHNMGADGIR",
        "modifications": ["15.99492:5"],
        "calc_mz": "1315.5700",
        "spectrum_title": "test_Creinhardtii_QE_pH11.10381.10381.3",
        "spectrum_id": "10381",
        "charge": "3",
        "retention_time_seconds": "1943.05878",
    }
    assert spec_records[70] == {
        "x!tandem:delta": "2.022",
        "x!tandem:hyperscore": "10.9",
        "x!tandem:nextscore": "8.0",
        "x!tandem:y_score": "5.0",
        "x!tandem:y_ions": "3",
        "x!tandem:b_score": "7.4",
        "x!tandem:b_ions": "1",
        "sequence": "FHVEALMHADLHFTRKHSR",
        "modifications": ["15.99492:6"],
        "calc_mz": "2348.183",
        "spectrum_title": "test_Creinhardtii_QE_pH11.41649.41649.2",
        "spectrum_id": "41649",
        "charge": "2",
        "retention_time_seconds": "5242.7709",
    }
    assert search_engine == "xtandem_alanine"



"""Engine parser."""

import xml.etree.ElementTree as etree

import pandas as pd
import regex as re
from loguru import logger

from pyiohat.parsers.ident_base_parser import IdentBaseParser


class XTandemAlanine_Parser(IdentBaseParser):
    """File parser for X!Tandem Alanine."""

    def __init__(self, *args, **kwargs):
        """Initialize parser.

        Reads in data file and provides mappings.
        """
        super().__init__(*args, **kwargs)
        self.spec_records = None
        self.search_engine = None
        self.style = "xtandem_style_1"
        self.mapping_dict = {
            v: k
            for k, v in self.param_mapper.get_default_params(style=self.style)[
                "header_translations"
            ]["translated_value"].items()
        }

    @classmethod
    def check_parser_compatibility(cls, file):
        """Assert compatibility between file and parser.

        Args:
            file (path object): path to input file

        Returns:
            bool: True if parser and file are compatible
        """
        is_xml = file.name.endswith(".xml")
        with open(file) as f:
            try:
                head = "".join([next(f) for _ in range(10)])
            except StopIteration:
                head = ""
        contains_ref = "tandem-style.xsl" in head

        return is_xml and contains_ref

    def get_spec_records(self):
        """Retrieve specs and search_engine from file.

        Returns:
            spec_records (list): information on PSMs
            search_engine (str): file version
        """
        spec_records = []
        search_engine = ""

        results = {}
        domains = []
        mods = []
        aa_mods = []

        for event, entry in etree.iterparse(self.input_file):
            entry_tag = entry.tag

            if entry_tag == ("aa"):
                aa_mods.append(
                    {"mass": entry.attrib["modified"], "at": entry.attrib["at"]}
                )
            elif entry_tag == ("domain"):
                for attrib in list(entry.attrib):
                    if attrib in self.mapping_dict:
                        _key = self.mapping_dict[attrib]
                        results[_key] = entry.attrib[attrib]
                for aa in aa_mods:
                    mass = aa["mass"]
                    at = aa["at"]
                    start = entry.attrib["start"]
                    rel_pos = int(at) - int(start)
                    mods.append(f"{mass}:{rel_pos}")
                results["modifications"] = mods
                results["calc_mz"] = entry.attrib["mh"]
                mods = []
                aa_mods = []
                domains.append(results)
                results = {}
            elif entry_tag == ("note"):
                if entry.attrib["label"] == "Description":
                    results["spectrum_title"] = entry.text.split()[0]
                    results["spectrum_id"] = entry.text.split(".")[-3]
                elif entry.attrib["label"] == "process, version":
                    search_engine = (
                        "xtandem_"
                        + re.search(r"(?<=Tandem )\w+", entry.text).group().lower()
                    )
            elif entry_tag == ("group"):
                if "id" in entry.attrib:
                    for attrib in list(entry.attrib):
                        if attrib in self.mapping_dict:
                            _key = self.mapping_dict[attrib]
                            results[_key] = entry.attrib[attrib]
                    for dom in domains:
                        dom.update(results)
                        spec_records.append(dom)
                    domains = []
                    results = {}
        return spec_records, search_engine

    def map_mod_names(self, df):
        """Map modifications on corresponding sequences.

        Args:
            df (pd.DataFrame): input dataframe

        Returns:
            df (pd.DataFrame): dataframe with processed modification column
        """
        unique_mods = set().union(*df["modifications"].apply(set).values)
        unique_mod_masses = {m.split(":")[0] for m in unique_mods}
        potential_names = {
            m: [
                name
                for name in self.mod_mapper.mass_to_names(float(m), decimals=4)
                if name in self.mod_dict
            ]
            for m in unique_mod_masses
        }
        mod_translation = {}
        new_mods = pd.Series("", index=df.index)
        for m in unique_mods:
            mass, pos = m.split(":")
            potential_mods = potential_names[mass]
            if len(potential_mods) == 0:
                mod_translation[m] = None
            else:
                for name in potential_mods:
                    # TODO: Is position 'any' respected here
                    in_seq = df["sequence"].str[int(pos)].isin(
                        self.mod_dict[name]["aa"]
                    ) & df["modifications"].str.join("|").str.contains(m)
                    n_term = (~in_seq) & (
                        (
                            ("N-term" in self.mod_dict[name]["position"])
                            | ("Prot-N-term" in self.mod_dict[name]["position"])
                        )
                        & df["modifications"].str.join("|").str.contains(m)
                    )
                    if in_seq.sum() != 0:
                        new_mods.loc[in_seq] += f"{name}:{int(pos)+1};"
                    elif n_term.sum() != 0:
                        new_mods.loc[n_term] += f"{name}:0;"
                    else:
                        logger.error(f"Modification {name} could not be mapped.")
                        raise KeyError
        df["modifications"] = new_mods.str.rstrip(";")

        return df

    def unify(self):
        """
        Primary method to read and unify engine output.

        Returns:
            self.df (pd.DataFrame): unified dataframe
        """
        self.spec_records, self.search_engine = self.get_spec_records()
        self.df = pd.DataFrame(self.spec_records)
        self.df["search_engine"] = self.search_engine
        self.df["calc_mz"] = (
            (self.df["calc_mz"].astype(float) - self.PROTON)
            / self.df["charge"].astype(int)
        ) + self.PROTON
        self.df = self.map_mod_names(self.df)
        self.process_unify_style()

        return self.df



"""Quant base parser class."""

from pyiohat.parsers.base_parser import BaseParser
import pandas as pd


class QuantBaseParser(BaseParser):
    """Base class of all quant parsers."""

    def __init__(self, *args, **kwargs):
        """Initialize parser.

        Reads in data file and provides mappings.
        """
        super().__init__(*args, **kwargs)
        self.required_headers = self._load_model("quant_parser_model.json")
        self.col_order = pd.Series(self.required_headers.keys())
        self.rt_lookup = self._read_meta_info_lookup_file()

    def process_unify_style(self):
        """Apply sanitizing methods."""
        # TODO: this should be called here too
        self.calculate_accuracy()
        self.sanitize()

    def calculate_accuracy(self):
        accuracy_mz = self.df.loc[
            self.df["reported_mz"] != "-", "theoretical_mz"
        ].astype(float) - self.df.loc[
            self.df["reported_mz"] != "-", "reported_mz"
        ].astype(
            float
        )
        accuracy_ppm = (
            (accuracy_mz)
            / self.df.loc[self.df["reported_mz"] != "-", "reported_mz"].astype(float)
        ) * 1e6
        self.df.loc[self.df["reported_mz"] != "-", "accuracy_ppm"] = accuracy_ppm
        self.df.loc[self.df["reported_mz"] == "-", "accuracy_ppm"] = -1
        self.df.loc[self.df["reported_mz"] == "-", "reported_ppm"] = -1

        self.df.loc[self.df["reported_mz"] != "-", "accuracy_mz"] = accuracy_mz
        self.df.loc[self.df["reported_mz"] == "-", "accuracy_mz"] = -1
        self.df.loc[self.df["reported_mz"] == "-", "reported_mz"] = -1



"""Engine parser."""
import multiprocessing as mp
import sys
from itertools import repeat

import numpy as np
import pandas as pd
import regex as re
from loguru import logger
from tqdm import tqdm

from pyiohat.parsers.ident_base_parser import IdentBaseParser
from pyiohat.utils import merge_and_join_dicts

mascot_custom_psm_regex = re.compile(
    r"(?:[-+0-9]+),(?P<exp_mass>[0-9\.]+),(?:[-0-9\.]+),(?P<n_matched_ions>[0-9]+),(?P<seq>[A-Z]+),(?:[0-9]+),(?P<opt_mod_string>[0-9]+),(?P<score>[.0-9]+),(?:[0-9]+),(?:.+subst;)(?P<subst>.+)"
)


def _get_single_spec_df(reference_dict, spectrum):
    """Primary method for reading and storing information from a single spectrum.

    Args:
        reference_dict (dict): dict with reference columns to be filled in
        spectrum (tuple): spectrum data

    Returns:
        (pd.DataFrame): dataframe for single spec id

    """
    spec_records = []
    spec_level_dict = reference_dict.copy()
    query, spec_level_info = spectrum[:2]

    spec_level_dict["spectrum_title"] = re.search(
        r"(?<=title=)(.+)", spec_level_info
    ).group()
    spec_level_dict["charge"] = re.search(r"(?<=charge=)(\d+)", spec_level_info).group()
    spec_level_dict["spectrum_id"] = re.search(
        r"(?<=scans=)(\d+)", spec_level_info
    ).group()
    spec_level_dict["retention_time_seconds"] = re.search(
        r"(?<=rtinseconds=)(\d+\.\d+)", spec_level_info
    ).group()

    # Iterate children
    for psm in spectrum[2]:
        psm_level_dict = spec_level_dict.copy()
        psm_level_info = re.search(mascot_custom_psm_regex, psm).groupdict()
        psm_level_dict["exp_mz"] = psm_level_info["exp_mass"]
        psm_level_dict["mascot:num_matched_ions"] = psm_level_info["n_matched_ions"]
        psm_level_dict["sequence"] = psm_level_info["seq"]
        psm_level_dict["modifications"] = psm_level_info["opt_mod_string"]
        psm_level_dict["mascot:score"] = psm_level_info["score"]
        psm_level_dict["subst"] = psm_level_info["subst"]

        spec_records.append(psm_level_dict)

    return pd.DataFrame(spec_records)


class Mascot_2_6_2_Parser(IdentBaseParser):
    """File parser for MSGF+."""

    def __init__(self, *args, **kwargs):
        """Initialize parser.

        Reads in data file and provides mappings.
        """
        super().__init__(*args, **kwargs)
        self.style = "mascot_style_1"

        self.section_data, self.spectrum_data = self._get_data_on_spectrum_level()
        self.mods = {
            "opt": dict(
                re.findall(r"delta([\d]+)=[\d.]+,(\S*)", self.section_data["masses"])
            ),
            "fix": dict(
                re.findall(
                    r"FixedMod[\d]+=[\d.]+,(\S*)\s\((\w)\)", self.section_data["masses"]
                )
            ),
        }

        self.reference_dict["search_engine"] = "mascot_" + re.search(
            r"(?<=version=).*", self.section_data["header"]
        ).group().replace(".", "_")
        self.reference_dict["mascot:score"] = pd.NA

    @classmethod
    def check_parser_compatibility(cls, file):
        """Assert compatibility between file and parser.

        Args:
            file (str): path to input file

        Returns:
            bool: True if parser and file are compatible

        """
        is_dat = file.as_posix().endswith(".dat")

        with open(file.as_posix()) as f:
            try:
                head = "".join([next(f) for _ in range(5)])
            except StopIteration:
                head = ""
        contains_engine = "Mascot" in head
        return is_dat and contains_engine

    def _get_data_on_spectrum_level(self):
        """Provide aggregated data on spectrum level."""
        with open(self.input_file) as f:
            file_str = f.read()

        file_section_pattern = re.compile(
            r"(?:Content-Type: application/x-Mascot; name=\")([\w+]*)"
        )
        section_split = re.split(file_section_pattern, file_str)[1:]
        section_data = {k: v for k, v in zip(section_split[::2], section_split[1::2])}

        # Filters for non empty data and only respective _subst metainfo
        filter_pattern = re.compile(r"q[\d]+_p[\d]+=(?!-1$).+|q[\d]+_p[\d]+_subst=.+")
        peptide_data = dict(
            [
                peptide.split("=")
                for peptide in section_data["peptides"].split("\n")[2:-2]
                if re.match(filter_pattern, peptide)
            ]
        )
        base_entries = {k: v for k, v in peptide_data.items() if "_subst" not in k}
        subst_entries = {
            k.rstrip("_subst"): v for k, v in peptide_data.items() if "_subst" in k
        }

        peptide_data = merge_and_join_dicts(
            [base_entries, subst_entries], delimiter=";subst;"
        )
        psm_info = pd.DataFrame(
            peptide_data.values(), index=peptide_data.keys(), columns=["info"]
        )
        psm_info.index = psm_info.index.str.replace("q", "query")
        psm_info.reset_index(inplace=True)
        psm_info.groupby("index")["info"].apply(list).to_dict()
        psm_info.loc[:, ["index", "psm"]] = (
            psm_info["index"].str.split("_", expand=True).values
        )
        psm_info = psm_info.groupby("index")["info"].apply(list).to_dict()

        spectrum_info = {k: v for k, v in section_data.items() if k in psm_info}
        spectrum_data = [
            (k, v_section, psm_info[k]) for k, v_section in spectrum_info.items()
        ]

        section_data = {k: v for k, v in section_data.items() if "query" not in k}

        return section_data, spectrum_data

    def _translate_opt_mods(self, raw_mod):
        """Replace internal modification nomenclature with formatted modification strings.

        Args:
            raw_mod (str): unformatted mod string

        Returns:
            formatted_mod_str (str): formatted mod string
        """
        formatted_mod_str = ";"
        for pos, mod_key in enumerate(raw_mod):
            if mod_key == "0":
                continue
            if mod_key in self.mods["opt"]:
                formatted_mod_str += f"{self.mods['opt'][mod_key]}:{pos};"
            else:
                logger.error(f"Modification {mod_key} could not be mapped.")
                raise KeyError

        return formatted_mod_str

    def _format_mods(self):
        """Convert mods to unified modstring.

        Operations are performed inplace.
        """
        fix_mods = None
        for name, aa in self.mods["fix"].items():
            fm_strings = (
                self.df["sequence"]
                .str.split(aa)
                .apply(
                    lambda l: ";".join(
                        [
                            name + ":" + ind
                            for ind in (
                                np.cumsum(list(map(len, l[:-1]))) + range(1, len(l))
                            ).astype(str)
                        ]
                    )
                )
            )
            if fix_mods is None:
                fix_mods = fm_strings
            else:
                fix_mods = fix_mods + ";" + fm_strings + ";"

        self.df.loc[:, "modifications"] = (
            self.df["modifications"].apply(self._translate_opt_mods).to_list()
        )
        if fix_mods is not None:
            self.df.loc[:, "modifications"] += fix_mods

        # Add substitutions
        if self.df["subst"].str.match(r"(\d+,\w,\w)").any():
            subst_df = pd.DataFrame(
                self.df["subst"].str.findall(r"(\d+,\w,\w)").tolist()
            )
            subst_df = (
                "Subst("
                + subst_df[0].str.split(",").str[1]
                + "):"
                + subst_df[0].str.split(",").str[0]
            ).fillna("")
            self.df.loc[:, "modifications"] += subst_df

        self.df.drop(columns="subst", inplace=True)

    def unify(self):
        """
        Primary method to read and unify engine output.

        Returns:
            self.df (pd.DataFrame): unified dataframe
        """
        logger.remove()
        logger.add(lambda msg: tqdm.write(msg, end=""))
        pbar_iterator = tqdm(
            zip(
                repeat(self.reference_dict),
                self.spectrum_data,
            ),
            total=len(self.spectrum_data),
        )
        with mp.Pool(self.params.get("cpus", mp.cpu_count() - 1)) as pool:
            chunk_dfs = pool.starmap(
                _get_single_spec_df,
                pbar_iterator,
            )
        logger.remove()
        logger.add(sys.stdout)
        self.df = pd.concat(chunk_dfs, axis=0, ignore_index=True)
        self.df.loc[:, "spectrum_title"] = (
            self.df["spectrum_title"]
            .str.replace("%2e", ".", regex=False)
            .str.replace(".mzML", "", regex=False)
        )
        self._format_mods()
        self.process_unify_style()

        return self.df



#!/usr/bin/env python
import pytest

from pyiohat.parsers.quant.tmt_quant_parser_1_0_0 import (
    TMTQuantParser,
)


def test_engine_parsers_tmt_quant_init():
    input_file = pytest._test_path / "data" / "mapping_data" / "trunc_tmt_quant.csv"
    rt_lookup_path = pytest._test_path / "data" / "TMT16_ursgal_lookup.csv"
    parser = TMTQuantParser(
        input_file,
        params={
            "rt_pickle_name": rt_lookup_path,
        },
    )


def test_engine_parsers_tmt_quant_check_parser_compatibility():
    input_file = pytest._test_path / "data" / "mapping_data" / "trunc_tmt_quant.csv"

    assert TMTQuantParser.check_parser_compatibility(input_file) is True


def test_engine_parsers_tmt_quant_file_not_matches_parser():
    input_file = (
        pytest._test_path / "data" / "test_Creinhardtii_QE_pH11_msfragger_3.tsv"
    )

    assert TMTQuantParser.check_parser_compatibility(input_file) is False


def test_engine_parsers_tmt_quant_unify_row():
    input_file = pytest._test_path / "data" / "mapping_data" / "trunc_tmt_quant.csv"
    rt_lookup_path = pytest._test_path / "data" / "TMT16_ursgal_lookup.csv"

    parser = TMTQuantParser(
        input_file,
        params={
            "rt_pickle_name": rt_lookup_path,
        },
    )
    df = parser.unify()
    assert len(df) == 7
    assert pytest.approx(df["retention_time_seconds"].mean()) == 781.94380
    assert pytest.approx(df["quant_value"].mean()) == 249596.16588
    assert pytest.approx(df["accuracy_mz"].mean()) == 0.009146615862846



"""Unify handler."""
from importlib import import_module
from pathlib import Path

import pandas as pd

from pyiohat.parsers.base_parser import BaseParser


class Unify:
    """Interface to unify ident outputs from different engines.

    Attributes:
        parser (`pyiohat.parsers.base_parser.__BaseParser`): Parser fitting the specified input_file

    """

    def __init__(self, input_file, params, immutable_peptides=None):
        """Initialize unifier.

        Args:
            input_file (str): path to input file
            params (dict): ursgal param dict
            immutable_peptides (str, optional): path to file with immutable peptides
        """
        if not isinstance(input_file, Path):
            self.input_file = Path(input_file)
        else:
            self.input_file = input_file
        self.params = params

        if immutable_peptides is not None:
            self.immutable_peptides = (
                pd.read_csv(immutable_peptides, header=None).iloc[:, 0].to_list()
            )
        else:
            self.immutable_peptides = None

        self._parser_classes = []
        self.parser = self._get_parser()
        self.df = None

    def _get_parser(self):
        """Check input file / parser compatibility and init matching parser in self.parser.

        Raises error if no matching parser can be found.
        """
        # Get all files except __init__.pys
        parser_files = (Path(__file__).parent / "parsers").rglob("[!_]*.py")
        # Make paths relative to package
        parser_files = [
            p.relative_to(Path(__file__).parent.parent).as_posix() for p in parser_files
        ]
        # Format and filter for relevant subdirectories
        parser_files = [
            p.replace(".py", "").replace("/", ".")
            for p in parser_files
            if p.count("/") == 3
        ]
        for parser in parser_files:
            import_module(parser)

        for cat in BaseParser.__subclasses__():
            self._parser_classes.extend(cat.__subclasses__())

        for parser in self._parser_classes:
            if parser.check_parser_compatibility(self.input_file) is True:
                return parser(
                    input_file=self.input_file,
                    params=self.params,
                    immutable_peptides=self.immutable_peptides,
                )

        raise IOError(f"No suitable parser found for {self.input_file}.")

    def get_dataframe(self):
        """Compute and returns a unified dataframe using the input file with specified parameters.

        Returns:
            self.df (pd.DataFrame): unified dataframe

        """
        self.df = self.parser.unify()

        return self.df



"""Quant parser."""
import re
from pathlib import Path

import pandas as pd
from chemical_composition import ChemicalComposition

from pyiohat.parsers.quant_base_parser import QuantBaseParser
from pyiohat.parsers.misc import get_compositions_and_monoisotopic_masses


class TMTQuantParser(QuantBaseParser):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.style = "cz_tmt_quant_style_1"
        self.df = pd.read_csv(self.input_file)

        self.mapping_dict = {
            v: k
            for k, v in self.param_mapper.get_default_params(style=self.style)[
                "header_translations"
            ]["translated_value"].items()
        }
        self.df.rename(columns=self.mapping_dict, inplace=True)

    @classmethod
    def check_parser_compatibility(cls, file):
        is_csv = file.as_posix().endswith(".csv")
        with open(file.as_posix(), "r") as fin:
            try:
                header = "".join([next(fin) for _ in range(1)])
            except StopIteration:
                header = ""
        header = set(header.rstrip("\n").split(","))

        ref_columns = {
            "filename",
            "mz",
            "raw_quant_intensity",
            "retention_time_seconds",
            "label",
            "mz_delta",
            "iso_mz",
            "ppm",
            "spectrum_id",
            "raw_quant_area",
            "isolabel_id",
            "original_quant_value",
            "s2i",
            "quant_value",
        }
        columns_match = len(ref_columns.difference(header)) == 0
        return columns_match and is_csv

    def unify(self):
        """
        Primary method to read and unify engine output.

        Returns:
            self.df (pd.DataFrame): unified dataframe
        """
        # base model
        self.df["charge"] = -1
        self.df["accuracy_ppm"] = -1
        self.df["accuracy_ppm_C12"] = -1
        self.df["chemical_composition"] = ""
        # quant model
        self.df["trivial_name"] = ""
        self.df["quant_score"] = -1
        self.df["quant_group"] = ""
        self.df["ccs"] = -1
        self.df["ident_reference"] = -1
        self.df["fwhm"] = -1
        self.df["quant_engine"] = "TMTQuant_1_0_0"
        self.process_unify_style()
        # required columns for recalculation not in dataframe
        self.df["accuracy_ppm"] = -1
        return self.df



"""Engine parser."""

import numpy as np
import pandas as pd

from pyiohat.parsers.ident_base_parser import IdentBaseParser


class Omssa_Parser(IdentBaseParser):
    """File parser for OMSSA."""

    def __init__(self, *args, **kwargs):
        """Initialize parser.

        Reads in data file and provides mappings.
        """
        super().__init__(*args, **kwargs)
        self.style = "omssa_style_1"

        self.df = pd.read_csv(self.input_file)

        self.mapping_dict = {
            v: k
            for k, v in self.param_mapper.get_default_params(style=self.style)[
                "header_translations"
            ]["translated_value"].items()
        }
        self.df.rename(columns=self.mapping_dict, inplace=True)
        self.df.columns = self.df.columns.str.lstrip(" ")
        self.df.drop(
            columns=[
                c
                for c in self.df.columns
                if c
                not in set(self.mapping_dict.values()) | set(self.reference_dict.keys())
            ],
            inplace=True,
            errors="ignore",
        )
        self.reference_dict.update({k: None for k in self.mapping_dict.values()})

    @classmethod
    def check_parser_compatibility(cls, file):
        """Assert compatibility between file and parser.

        Args:
            file (str): path to input file

        Returns:
            bool: True if parser and file are compatible

        """
        is_csv = file.as_posix().endswith(".csv")
        with open(file.as_posix()) as f:
            try:
                head = "".join([next(f) for _ in range(1)])
            except StopIteration:
                head = ""
        head = set(head.rstrip("\n").split(","))
        ref_columns = {
            "Spectrum number",
            " Filename/id",
            " Peptide",
            " E-value",
            " Mass",
            " gi",
            " Accession",
            " Start",
            " Stop",
            " Defline",
            " Mods",
            " Charge",
            " Theo Mass",
            " P-value",
            " NIST score",
        }
        columns_match = len(ref_columns.difference(head)) == 0
        return is_csv and columns_match

    def _replace_mod_strings(self, row, mod_translations):
        """Replace single mod string.

        Args:
            row (str): unprocessed modification string
            mod_translations (dict): mod translation dict

        Returns:
            mod_str (str): formatted modification string
        """
        if row == "":
            return ""
        mods = row.split(" ,")
        new_modstring = ""
        for m in mods:
            omssa_name, pos = m.split(":")
            unimod_name = mod_translations[omssa_name]["unimod_name"]
            if pos == "1" and any(
                [
                    ("N-term" in target)
                    for target in mod_translations[omssa_name]["aa_targets"]
                ]
            ):
                pos = "0"
            new_modstring += f"{unimod_name}:{pos};"
        return new_modstring.rstrip(";")

    def translate_mods(self):
        """
        Replace internal modification nomenclature with formatted modification strings.

        Operations are performed inplace.
        """
        self.df["sequence"] = self.df["sequence"].str.upper()
        self.df["modifications"] = (
            self.df["modifications"].fillna("").str.replace(" ,", ";")
        )
        fix_mods = None
        # Map fixed mods
        fixed_mod_types = [
            d for d in self.params["modifications"] if d["type"] == "fix"
        ]
        for fm in fixed_mod_types:
            fm_strings = (
                self.df["sequence"]
                .str.split(fm["aa"])
                .apply(
                    lambda l: ";".join(
                        [
                            fm["name"] + ":" + ind
                            for ind in (
                                np.cumsum(list(map(len, l[:-1]))) + range(1, len(l))
                            ).astype(str)
                        ]
                    )
                )
            )
            if fix_mods is None:
                fix_mods = fm_strings
            else:
                fix_mods = fix_mods + ";" + fm_strings

        if fix_mods is not None:
            self.df["modifications"] = (
                self.df["modifications"].fillna("") + ";" + fix_mods
            )

    def unify(self):
        """
        Primary method to read and unify engine output.

        Returns:
            self.df (pd.DataFrame): unified dataframe
        """
        self.df["spectrum_id"] = (
            self.df["spectrum_title"].str.split(".").str[-3].astype(int)
        )
        self.df["search_engine"] = "omssa_2_1_9"
        self.translate_mods()
        self.process_unify_style()

        return self.df



#!/usr/bin/env python

import pytest

from pyiohat.parsers.ident.omssa_2_1_9_parser import Omssa_Parser


def test_engine_parsers_omssa_init():
    input_file = (
        pytest._test_path / "data" / "test_Creinhardtii_QE_pH11_omssa_2_1_9.csv"
    )

    parser = Omssa_Parser(
        input_file,
        params={
            "cpus": 2,
            "enzyme": "(?<=[KR])(?![P])",
            "terminal_cleavage_site_integrity": "any",
            "validation_score_field": {"omssa_2_1_9": "omssa:pvalue"},
            "bigger_scores_better": {"omssa_2_1_9": False},
            "modifications": [
                {
                    "aa": "M",
                    "type": "opt",
                    "position": "any",
                    "name": "Oxidation",
                },
                {
                    "aa": "C",
                    "type": "fix",
                    "position": "any",
                    "name": "Carbamidomethyl",
                },
                {
                    "aa": "*",
                    "type": "opt",
                    "position": "Prot-N-term",
                    "name": "Acetyl",
                },
            ],
            "xml_file_list": [
                pytest._test_path / "data" / "mods.xml",
                pytest._test_path / "data" / "usermods.xml",
            ],
        },
    )


def test_engine_parsers_omssa_check_parser_compatibility():
    input_file = (
        pytest._test_path / "data" / "test_Creinhardtii_QE_pH11_omssa_2_1_9.csv"
    )
    assert Omssa_Parser.check_parser_compatibility(input_file) is True


def test_engine_parsers_omssa_check_dataframe_integrity():
    input_file = (
        pytest._test_path / "data" / "test_Creinhardtii_QE_pH11_omssa_2_1_9.csv"
    )
    rt_lookup_path = pytest._test_path / "data" / "_ursgal_lookup.csv"
    db_path = pytest._test_path / "data" / "test_Creinhardtii_target_decoy.fasta"

    parser = Omssa_Parser(
        input_file,
        params={
            "cpus": 2,
            "rt_pickle_name": rt_lookup_path,
            "database": db_path,
            "enzyme": "(?<=[KR])(?![P])",
            "terminal_cleavage_site_integrity": "any",
            "validation_score_field": {"omssa_2_1_9": "omssa:pvalue"},
            "bigger_scores_better": {"omssa_2_1_9": False},
            "modifications": [
                {
                    "aa": "M",
                    "type": "opt",
                    "position": "any",
                    "name": "Oxidation",
                },
                {
                    "aa": "C",
                    "type": "fix",
                    "position": "any",
                    "name": "Carbamidomethyl",
                },
                {
                    "aa": "*",
                    "type": "opt",
                    "position": "Prot-N-term",
                    "name": "Acetyl",
                },
            ],
            "xml_file_list": [
                pytest._test_path / "data" / "mods.xml",
                pytest._test_path / "data" / "usermods.xml",
            ],
        },
    )
    df = parser.unify()
    assert pytest.approx(df["ucalc_mz"].mean()) == 826.67596
    assert (df["raw_data_location"] == "path/for/glory.mzML").all()
    assert pytest.approx(df["exp_mz"].mean()) == 826.7788

    assert df["modifications"].str.contains("Acetyl:0").sum() == 5
    assert df["modifications"].str.contains("Oxidation:").sum() == 93
    assert (
        df["modifications"].str.count("Carbamidomethyl:")
        == df["sequence"].str.count("C")
    ).all()
    assert df["modifications"].str.count(":").sum() == 204


def test_replace_mod_strings_omssa():
    input_file = (
        pytest._test_path / "data" / "test_Creinhardtii_QE_pH11_omssa_2_1_9.csv"
    )

    parser = Omssa_Parser(
        input_file,
        params={
            "cpus": 2,
            "enzyme": "(?<=[KR])(?![P])",
            "terminal_cleavage_site_integrity": "any",
            "validation_score_field": {"omssa_2_1_9": "omssa:pvalue"},
            "bigger_scores_better": {"omssa_2_1_9": False},
            "modifications": [
                {
                    "aa": "M",
                    "type": "opt",
                    "position": "any",
                    "name": "Oxidation",
                },
                {
                    "aa": "C",
                    "type": "fix",
                    "position": "any",
                    "name": "Carbamidomethyl",
                },
                {
                    "aa": "*",
                    "type": "opt",
                    "position": "Prot-N-term",
                    "name": "Acetyl",
                },
            ],
            "xml_file_list": [
                pytest._test_path / "data" / "mods.xml",
                pytest._test_path / "data" / "usermods.xml",
            ],
        },
    )
    raw_mod = "oxidation of M:4"
    converted = parser._replace_mod_strings(
        raw_mod,
        {
            "acetylation of protein n-term": {
                "unimod_id": None,
                "omssa_unimod_id": "1",
                "unimod_name": "Acetyl",
                "omssa_name": "acetylation of protein n-term",
                "aa_targets": [],
            },
            "oxidation of M": {
                "unimod_id": None,
                "omssa_unimod_id": "35",
                "unimod_name": "Oxidation",
                "omssa_name": "oxidation of M",
                "aa_targets": ["M"],
            },
        },
    )
    assert converted == "Oxidation:4"



#!/usr/bin/env python
from pathlib import Path

import pytest

import pyiohat
from pyiohat.parsers.ident.comet_2020_01_4_parser import (
    Comet_2020_01_4_Parser,
)
from pyiohat.parsers.ident.mascot_2_6_2_parser import Mascot_2_6_2_Parser
from pyiohat.parsers.ident.msamanda_2_parser import MSAmanda_2_Parser
from pyiohat.parsers.ident.msfragger_3_parser import MSFragger_3_Parser
from pyiohat.parsers.ident.msgfplus_2021_03_22_parser import (
    MSGFPlus_2021_03_22_Parser,
)
from pyiohat.parsers.ident.omssa_2_1_9_parser import Omssa_Parser
from pyiohat.parsers.ident.xtandem_alanine import XTandemAlanine_Parser
from pyiohat.unify import Unify


def test_unify_get_parser_classes():
    rt_lookup_path = pytest._test_path / "data" / "_ursgal_lookup.csv"
    p = pytest._test_path / "data" / "test_Creinhardtii_QE_pH11_omssa_2_1_9.csv"
    db_path = pytest._test_path / "data" / "BSA.fasta"
    u = Unify(
        p,
        {
            "rt_pickle_name": rt_lookup_path,
            "database": db_path,
            "modifications": [
                {
                    "aa": "M",
                    "type": "opt",
                    "position": "any",
                    "name": "Oxidation",
                },
                {
                    "aa": "C",
                    "type": "fix",
                    "position": "any",
                    "name": "Carbamidomethyl",
                },
                {
                    "aa": "*",
                    "type": "opt",
                    "position": "Prot-N-term",
                    "name": "Acetyl",
                },
            ],
            "xml_file_list": [
                pytest._test_path / "data" / "mods.xml",
                pytest._test_path / "data" / "usermods.xml",
            ],
        },
    )
    # Get files and subtract __init__.py
    ident_files = (
        len(list((Path(pyiohat.__path__[0]) / "parsers" / "ident").glob("*.py"))) - 1
    )
    quant_files = (
        len(list((Path(pyiohat.__path__[0]) / "parsers" / "quant").glob("*.py"))) - 1
    )
    assert len(u._parser_classes) == ident_files + quant_files


def test_unify_get_omssa_parser():
    rt_lookup_path = pytest._test_path / "data" / "_ursgal_lookup.csv"
    p = pytest._test_path / "data" / "test_Creinhardtii_QE_pH11_omssa_2_1_9.csv"
    db_path = pytest._test_path / "data" / "BSA.fasta"
    u = Unify(
        p,
        {
            "rt_pickle_name": rt_lookup_path,
            "database": db_path,
            "modifications": [
                {
                    "aa": "M",
                    "type": "opt",
                    "position": "any",
                    "name": "Oxidation",
                },
                {
                    "aa": "C",
                    "type": "fix",
                    "position": "any",
                    "name": "Carbamidomethyl",
                },
                {
                    "aa": "*",
                    "type": "opt",
                    "position": "Prot-N-term",
                    "name": "Acetyl",
                },
            ],
            "xml_file_list": [
                pytest._test_path / "data" / "mods.xml",
                pytest._test_path / "data" / "usermods.xml",
            ],
        },
    )
    assert isinstance(u.parser, Omssa_Parser)


def test_unify_get_msgfplus_parser():
    rt_lookup_path = pytest._test_path / "data" / "_ursgal_lookup.csv"
    p = pytest._test_path / "data" / "BSA1_msgfplus_2021_03_22.mzid"
    db_path = pytest._test_path / "data" / "BSA.fasta"
    u = Unify(
        p,
        {
            "rt_pickle_name": rt_lookup_path,
            "database": db_path,
            "modifications": [
                {
                    "aa": "M",
                    "type": "opt",
                    "position": "any",
                    "name": "Oxidation",
                },
                {
                    "aa": "C",
                    "type": "fix",
                    "position": "any",
                    "name": "Carbamidomethyl",
                },
                {
                    "aa": "*",
                    "type": "opt",
                    "position": "Prot-N-term",
                    "name": "Acetyl",
                },
            ],
        },
    )
    assert isinstance(u.parser, MSGFPlus_2021_03_22_Parser)


def test_unify_get_msamanda_parser():
    rt_lookup_path = pytest._test_path / "data" / "BSA1_ursgal_lookup.csv"
    p = pytest._test_path / "data" / "BSA1_msamanda_2_0_0_17442.csv"
    db_path = pytest._test_path / "data" / "BSA.fasta"
    u = Unify(
        p,
        {
            "rt_pickle_name": rt_lookup_path,
            "database": db_path,
            "modifications": [
                {
                    "aa": "M",
                    "type": "opt",
                    "position": "any",
                    "name": "Oxidation",
                },
                {
                    "aa": "C",
                    "type": "fix",
                    "position": "any",
                    "name": "Carbamidomethyl",
                },
                {
                    "aa": "*",
                    "type": "opt",
                    "position": "Prot-N-term",
                    "name": "Acetyl",
                },
            ],
        },
    )
    assert isinstance(u.parser, MSAmanda_2_Parser)


def test_unify_get_xtandem_parser():
    rt_lookup_path = pytest._test_path / "data" / "_ursgal_lookup.csv"
    p = pytest._test_path / "data" / "BSA1_xtandem_alanine.xml"
    db_path = pytest._test_path / "data" / "BSA.fasta"
    u = Unify(
        p,
        {
            "rt_pickle_name": rt_lookup_path,
            "database": db_path,
            "modifications": [
                {
                    "aa": "M",
                    "type": "opt",
                    "position": "any",
                    "name": "Oxidation",
                },
                {
                    "aa": "C",
                    "type": "fix",
                    "position": "any",
                    "name": "Carbamidomethyl",
                },
                {
                    "aa": "*",
                    "type": "opt",
                    "position": "Prot-N-term",
                    "name": "Acetyl",
                },
            ],
        },
    )
    assert isinstance(u.parser, XTandemAlanine_Parser)


def test_unify_get_msfragger_parser():
    rt_lookup_path = pytest._test_path / "data" / "_ursgal_lookup.csv"
    p = pytest._test_path / "data" / "test_Creinhardtii_QE_pH11_msfragger_3.tsv"
    db_path = pytest._test_path / "data" / "BSA.fasta"
    u = Unify(
        p,
        {
            "rt_pickle_name": rt_lookup_path,
            "database": db_path,
            "modifications": [
                {
                    "aa": "M",
                    "type": "opt",
                    "position": "any",
                    "name": "Oxidation",
                },
                {
                    "aa": "C",
                    "type": "fix",
                    "position": "any",
                    "name": "Carbamidomethyl",
                },
                {
                    "aa": "*",
                    "type": "opt",
                    "position": "Prot-N-term",
                    "name": "Acetyl",
                },
            ],
        },
    )
    assert isinstance(u.parser, MSFragger_3_Parser)


def test_unify_get_comet_parser():
    rt_lookup_path = pytest._test_path / "data" / "_ursgal_lookup.csv"
    p = pytest._test_path / "data" / "BSA1_comet_2020_01_4.mzid"
    db_path = pytest._test_path / "data" / "BSA.fasta"
    u = Unify(
        p,
        {
            "rt_pickle_name": rt_lookup_path,
            "database": db_path,
            "modifications": [
                {
                    "aa": "M",
                    "type": "opt",
                    "position": "any",
                    "name": "Oxidation",
                },
                {
                    "aa": "C",
                    "type": "fix",
                    "position": "any",
                    "name": "Carbamidomethyl",
                },
                {
                    "aa": "*",
                    "type": "opt",
                    "position": "Prot-N-term",
                    "name": "Acetyl",
                },
            ],
        },
    )
    assert isinstance(u.parser, Comet_2020_01_4_Parser)


def test_unify_get_mascot_parser():
    rt_lookup_path = pytest._test_path / "data" / "_ursgal_lookup.csv"
    p = pytest._test_path / "data" / "BSA1_mascot_2_6_2.dat"
    db_path = pytest._test_path / "data" / "BSA.fasta"
    u = Unify(
        p,
        {
            "rt_pickle_name": rt_lookup_path,
            "database": db_path,
            "modifications": [
                {
                    "aa": "M",
                    "type": "opt",
                    "position": "any",
                    "name": "Oxidation",
                },
                {
                    "aa": "C",
                    "type": "fix",
                    "position": "any",
                    "name": "Carbamidomethyl",
                },
                {
                    "aa": "*",
                    "type": "opt",
                    "position": "Prot-N-term",
                    "name": "Acetyl",
                },
            ],
        },
    )
    assert isinstance(u.parser, Mascot_2_6_2_Parser)



"""Dummy parser."""
from pyiohat.parsers.ident_base_parser import IdentBaseParser


class Dummy(IdentBaseParser):
    """File parser for Dummy."""

    def __init__(self, *args, **kwargs):
        """Initialize parser.

        Reads in data file and provides mappings.
        """
        super().__init__(*args, **kwargs)
        print("Miss Hoover, I glued my head to my shoulders.")

    @classmethod
    def check_parser_compatibility(cls, file):
        """Assert compatibility between file and parser.

        Args:
            file (str): path to input file

        Returns:
            bool: True if parser and file are compatible

        """
        return False

    def unify(self):
        """Primary method to read and unify engine output.

        Returns:
            self.df (pd.DataFrame): unified dataframe
        """
        return None



#!/usr/bin/env python

import pytest

from pyiohat.parsers.ident.msamanda_2_parser import MSAmanda_2_Parser


def test_engine_parsers_msamanda_init():
    input_file = pytest._test_path / "data" / "BSA1_msamanda_2_0_0_17442.csv"

    parser = MSAmanda_2_Parser(
        input_file,
        params={
            "cpus": 2,
            "enzyme": "(?<=[KR])(?![P])",
            "terminal_cleavage_site_integrity": "any",
            "validation_score_field": {"msamanda_2_0_0_17442": "amanda:score"},
            "bigger_scores_better": {"msamanda_2_0_0_17442": True},
            "modifications": [
                {
                    "aa": "M",
                    "type": "opt",
                    "position": "any",
                    "name": "Oxidation",
                },
                {
                    "aa": "C",
                    "type": "fix",
                    "position": "any",
                    "name": "Carbamidomethyl",
                },
                {
                    "aa": "*",
                    "type": "opt",
                    "position": "Prot-N-term",
                    "name": "Acetyl",
                },
            ],
        },
    )


def test_engine_parsers_msamanda_check_parser_compatibility():
    input_file = pytest._test_path / "data" / "BSA1_msamanda_2_0_0_17442.csv"
    assert MSAmanda_2_Parser.check_parser_compatibility(input_file) is True


def test_engine_parsers_msamanda_check_dataframe_integrity():
    input_file = pytest._test_path / "data" / "BSA1_msamanda_2_0_0_17442.csv"
    rt_lookup_path = pytest._test_path / "data" / "BSA1_ursgal_lookup.csv"
    db_path = pytest._test_path / "data" / "BSA.fasta"

    parser = MSAmanda_2_Parser(
        input_file,
        params={
            "cpus": 2,
            "rt_pickle_name": rt_lookup_path,
            "database": db_path,
            "enzyme": "(?<=[KR])(?![P])",
            "terminal_cleavage_site_integrity": "any",
            "validation_score_field": {"msamanda_2_0_0_17442": "amanda:score"},
            "bigger_scores_better": {"msamanda_2_0_0_17442": True},
            "modifications": [
                {
                    "aa": "M",
                    "type": "opt",
                    "position": "any",
                    "name": "Oxidation",
                },
                {
                    "aa": "C",
                    "type": "fix",
                    "position": "any",
                    "name": "Carbamidomethyl",
                },
                {
                    "aa": "*",
                    "type": "opt",
                    "position": "Prot-N-term",
                    "name": "Acetyl",
                },
            ],
        },
    )
    df = parser.unify()
    assert len(df) == 87
    assert pytest.approx(df["ucalc_mz"].mean()) == 485.2679
    assert pytest.approx(df["exp_mz"].mean()) == 485.26797

    assert df["modifications"].str.contains("Acetyl:0").sum() == 0
    assert (
        df["modifications"].str.count("Carbamidomethyl:")
        == df["sequence"].str.count("C")
    ).all()
    assert df["modifications"].str.count(":").sum() == 66
    assert (df["raw_data_location"] == "path/for/glory.mzML").all()


def test_map_mod_translation_msamanda():
    input_file = pytest._test_path / "data" / "BSA1_msamanda_2_0_0_17442.csv"

    parser = MSAmanda_2_Parser(
        input_file,
        params={
            "cpus": 2,
            "enzyme": "(?<=[KR])(?![P])",
            "terminal_cleavage_site_integrity": "any",
            "validation_score_field": {"msamanda_2_0_0_17442": "amanda:score"},
            "bigger_scores_better": {"msamanda_2_0_0_17442": True},
            "modifications": [
                {
                    "aa": "M",
                    "type": "opt",
                    "position": "any",
                    "name": "Oxidation",
                },
                {
                    "aa": "C",
                    "type": "fix",
                    "position": "any",
                    "name": "Carbamidomethyl",
                },
                {
                    "aa": "*",
                    "type": "opt",
                    "position": "Prot-N-term",
                    "name": "Acetyl",
                },
            ],
        },
    )
    converted = parser._map_mod_translation(row=["C3(Carbamidomethyl|57.021464|fixed)"])
    assert converted == "Carbamidomethyl:3;"



#!/usr/bin/env python
import pytest

from pyiohat.parsers.ident.comet_2020_01_4_parser import (
    Comet_2020_01_4_Parser,
)


def test_engine_parsers_comet_init():
    input_file = pytest._test_path / "data" / "BSA1_comet_2020_01_4.mzid"

    parser = Comet_2020_01_4_Parser(
        input_file,
        params={
            "cpus": 2,
            "modifications": [
                {
                    "aa": "M",
                    "type": "opt",
                    "position": "any",
                    "name": "Oxidation",
                },
                {
                    "aa": "C",
                    "type": "fix",
                    "position": "any",
                    "name": "Carbamidomethyl",
                },
                {
                    "aa": "*",
                    "type": "opt",
                    "position": "Prot-N-term",
                    "name": "Acetyl",
                },
            ],
        },
    )


def test_engine_parsers_comet_check_parser_compatibility():
    msgf_parser_class = Comet_2020_01_4_Parser
    input_file = pytest._test_path / "data" / "BSA1_comet_2020_01_4.mzid"
    assert msgf_parser_class.check_parser_compatibility(input_file) is True


def test_engine_parsers_comet_check_parser_compatibility_fail_with_omssa_file():
    msgf_parser_class = Comet_2020_01_4_Parser
    input_file = (
        pytest._test_path / "data" / "test_Creinhardtii_QE_pH11_omssa_2_1_9.csv"
    )
    assert msgf_parser_class.check_parser_compatibility(input_file) is False


def test_engine_parsers_comet_check_dataframe_integrity():
    input_file = pytest._test_path / "data" / "BSA1_comet_2020_01_4.mzid"
    rt_lookup_path = pytest._test_path / "data" / "BSA1_ursgal_lookup.csv"
    db_path = pytest._test_path / "data" / "BSA.fasta"

    parser = Comet_2020_01_4_Parser(
        input_file,
        params={
            "cpus": 2,
            "enzyme": "(?<=[KR])(?![P])",
            "terminal_cleavage_site_integrity": "any",
            "validation_score_field": {"comet_2020_01_4": "comet:e_value"},
            "bigger_scores_better": {"comet_2020_01_4": False},
            "rt_pickle_name": rt_lookup_path,
            "database": db_path,
            "modifications": [
                {
                    "aa": "M",
                    "type": "opt",
                    "position": "any",
                    "name": "Oxidation",
                },
                {
                    "aa": "C",
                    "type": "fix",
                    "position": "any",
                    "name": "Carbamidomethyl",
                },
                {
                    "aa": "*",
                    "type": "opt",
                    "position": "Prot-N-term",
                    "name": "Acetyl",
                },
            ],
        },
    )
    df = parser.unify()
    assert pytest.approx(df["ucalc_mz"].mean()) == 457.85944
    assert pytest.approx(df["exp_mz"].mean()) == 457.87625

    assert df["modifications"].str.contains("Acetyl:0").sum() == 5
    assert df["modifications"].str.contains("Oxidation:").sum() == 0
    assert (
        df["modifications"].str.count("Carbamidomethyl:")
        == df["sequence"].str.count("C")
    ).all()
    assert df["modifications"].str.count(":").sum() == 38
    assert (df["raw_data_location"] == "path/for/glory.mzML").all()


def test_engine_parsers_comet_get_version():
    input_file = pytest._test_path / "data" / "BSA1_comet_2020_01_4.mzid"
    parser = Comet_2020_01_4_Parser(input_file, params=None)
    version = parser.get_version()
    assert version == "comet_2020_01_4"


def test_engine_parser_comet_map_mod_mass():
    input_file = pytest._test_path / "data" / "BSA1_comet_2020_01_4.mzid"
    parser = Comet_2020_01_4_Parser(input_file=input_file, params=None)
    fixed_mods, modifications = parser.map_mod_mass()
    assert fixed_mods == {"C": "Carbamidomethyl"}
    assert modifications == {
        "57.021464": "Carbamidomethyl",
        "15.994915": "Oxidation",
        "42.010565": "Acetyl",
    }


def test_engine_parsers_comet_get_peptide_lookup():
    input_file = pytest._test_path / "data" / "BSA1_comet_2020_01_4.mzid"
    parser = Comet_2020_01_4_Parser(input_file=input_file, params=None)
    parser.fixed_mods, parser.mod_mass_map = parser.map_mod_mass()
    peptide_lookup = parser.get_peptide_lookup()
    assert len(peptide_lookup) == 24
    assert "EACFAVEGPK;10:42.010565;" in peptide_lookup.keys()
    assert peptide_lookup["LVTDLTK;"] == ("", "LVTDLTK")


def test_engine_parsers_comet_get_spec_records():
    input_file = pytest._test_path / "data" / "BSA1_comet_2020_01_4.mzid"
    parser = Comet_2020_01_4_Parser(input_file=input_file, params=None)
    parser.fixed_mods, parser.mod_mass_map = parser.map_mod_mass()
    parser.peptide_lookup = parser.get_peptide_lookup()
    spec_records = parser.get_spec_records()

    assert len(spec_records) == 60
    assert spec_records[4] == {
        "comet:num_matched_ions": "3",
        "comet:num_unmatched_ions": "33",
        "comet:xcorr": "0.3634",
        "comet:deltacn": "1.0000",
        "comet:score": "6.4000",
        "comet:e_value": "3.35E+01",
        "charge": "3",
        "sequence": "ECCDKPLLEK",
        "exp_mz": "431.205597",
        "calc_mz": "431.205546",
        "modifications": "Carbamidomethyl:2;Carbamidomethyl:3",
        "spectrum_id": "2569",
    }
    assert spec_records[58] == {
        "comet:num_matched_ions": "2",
        "comet:num_unmatched_ions": "50",
        "comet:xcorr": "0.4654",
        "comet:deltacn": "1.0000",
        "comet:score": "5.7000",
        "comet:e_value": "2.96E+01",
        "charge": "3",
        "sequence": "VPQVSTPTLVEVSR",
        "exp_mz": "504.618805",
        "calc_mz": "504.619112",
        "modifications": "",
        "spectrum_id": "3547",
    }


def test_engine_parsers_comet_map_mods_sequences():
    parser = Comet_2020_01_4_Parser(input_file=None, params=None)
    parser.fixed_mods = {"C": "Carbamidomethyl", "O": "Oxidation"}
    sequence = "ABCDEFCABC"
    mods = parser.map_mods_sequences(sequence)
    assert mods == "Carbamidomethyl:3;Carbamidomethyl:7;Carbamidomethyl:10"

    sequence = "ABCDOEFOCABC"
    mods = parser.map_mods_sequences(sequence)
    assert (
        mods
        == "Carbamidomethyl:3;Carbamidomethyl:9;Carbamidomethyl:12;Oxidation:5;Oxidation:8"
    )

    sequence = "OOOOOCCCCCMMMMM"
    mods = parser.map_mods_sequences(sequence)
    assert (
        mods
        == "Carbamidomethyl:6;Carbamidomethyl:7;Carbamidomethyl:8;Carbamidomethyl:9;Carbamidomethyl:10;Oxidation:1;Oxidation:2;Oxidation:3;Oxidation:4;Oxidation:5"
    )

    sequence = "CD"
    mods = parser.map_mods_sequences(sequence)
    assert mods == "Carbamidomethyl:1"

    sequence = "C"
    mods = parser.map_mods_sequences(sequence)
    assert mods == "Carbamidomethyl:1"

    parser.fixed_mods = {}
    sequence = "ASLDPOCSADK"
    with pytest.raises(ValueError):
        parser.map_mods_sequences(sequence)



import pytest

import pyiohat

COMET_TEST_FILE = pytest._test_path / "data" / "mapping_data" / "trunc_comet.mzid"
# MASCOT_TEST_FILE = ""
# MSAMANDA_TEST_FILE = ""
MSFRAGGER_TEST_FILE = (
    pytest._test_path / "data" / "mapping_data" / "trunc_msfragger.tsv"
)
MSGFPLUS_TEST_FILE = pytest._test_path / "data" / "mapping_data" / "trunc_msgfplus.mzid"
OMSSA_TEST_FILE = pytest._test_path / "data" / "mapping_data" / "trunc_omssa.csv"
XTANDEM_TEST_FILE = pytest._test_path / "data" / "mapping_data" / "trunc_xtandem.xml"

PARAMS = {
    "database": pytest._test_path / "data" / "mapping_data" / "trunc_fasta.protein.faa",
    "rt_pickle_name": pytest._test_path
    / "data"
    / "mapping_data"
    / "trunc_meta.spectra_meta.csv",
    "enzyme": "(?<=[KR])(?![P])",
    "terminal_cleavage_site_integrity": "any",
    "validation_score_field": {
        "msfragger_3_0": "msfragger:hyperscore",
        "comet_2020_01_4": "comet:e_value",
        "msgfplus_2021_03_22": "ms-gf:spec_evalue",
        "xtandem_alanine": "x!tandem:hyperscore",
        "omssa_2_1_9": "omssa:pvalue",
    },
    "bigger_scores_better": {
        "msfragger_3_0": True,
        "comet_2020_01_4": False,
        "msgfplus_2021_03_22": False,
        "xtandem_alanine": True,
        "omssa_2_1_9": True,
    },
    "modifications": [
        {
            "aa": "M",
            "type": "opt",
            "position": "any",
            "name": "Oxidation",
        },
        {
            "aa": "C",
            "type": "fix",
            "position": "any",
            "name": "Carbamidomethyl",
        },
        {
            "aa": "*",
            "type": "opt",
            "position": "Prot-N-term",
            "name": "Acetyl",
        },
        {
            "aa": "K",
            "type": "fix",
            "position": "any",
            "name": "TMTpro",
        },
        {
            "aa": "*",
            "type": "opt",
            "position": "N-term",
            "name": "TMTpro",
        },
    ],
}


def test_comet_mapping():
    df = pyiohat.Unify(input_file=COMET_TEST_FILE, params=PARAMS).get_dataframe()
    mod_str = df.loc[0, "modifications"]
    assert mod_str == "TMTpro:0;Carbamidomethyl:8;Oxidation:12;Carbamidomethyl:19"


# def test_mascot_mapping():
#     df = pd.read_csv(MASCOT_TEST_FILE)
#     assert df["modifications"].str.contains("TMTpro").all() is True
#     assert df["modifications"].str.contains("Carbamidomethyl").all() is True
#
#
# def test_msamanda_mapping():
#     df = pd.read_csv(MSAMANDA_TEST_FILE)
#     assert df["modifications"].str.contains("TMTpro").all() is True
#     assert df["modifications"].str.contains("Carbamidomethyl").all() is True


def test_msfragger_mapping():
    df = pyiohat.Unify(input_file=MSFRAGGER_TEST_FILE, params=PARAMS).get_dataframe()
    mod_str = df.loc[0, "modifications"]
    assert mod_str == "TMTpro:0;Carbamidomethyl:6"


def test_msgfplus_mapping():
    df = pyiohat.Unify(input_file=MSGFPLUS_TEST_FILE, params=PARAMS).get_dataframe()
    mod_str = df.loc[0, "modifications"]
    assert mod_str == "TMTpro:0;Oxidation:7;Carbamidomethyl:14;TMTpro:26;TMTpro:27"


def test_omssa_mapping():
    df = pyiohat.Unify(input_file=OMSSA_TEST_FILE, params=PARAMS).get_dataframe()
    mod_str = df.loc[0, "modifications"]
    assert (
        mod_str == "TMTpro:0;Carbamidomethyl:5;Carbamidomethyl:17;TMTpro:20;TMTpro:27"
    )


def test_xtandem_mapping():
    df = pyiohat.Unify(input_file=XTANDEM_TEST_FILE, params=PARAMS).get_dataframe()
    mod_str = df.loc[0, "modifications"]
    assert mod_str == "TMTpro:0;Oxidation:6;Carbamidomethyl:7"



"""Collection of utils."""


def merge_and_join_dicts(list_of_dicts, delimiter):
    """Merge list of dicts with identical keys as strings into single merged dict.

    Args:
        list_of_dicts (list): list of dicts which are to be merged
        delimiter (str): delimiter

    Returns:
        dict: preserved original keys with all values merged as strings with delimiter

    """
    return {
        key: delimiter.join([str(d.get(key)) for d in list_of_dicts])
        for key in set().union(*list_of_dicts)
    }



"""Parser handler."""
import csv
import json
from pathlib import Path

import numpy as np
import pandas as pd
import uparma
from chemical_composition import ChemicalComposition
from loguru import logger
from unimod_mapper.unimod_mapper import UnimodMapper


class BaseParser:
    """Base class of all parser types."""

    def __init__(self, input_file, params, immutable_peptides=None):
        """Initialize parser.

        Args:
            input_file (str): path to input file
            params (dict): ursgal param dict
            immutable_peptides (list, optional): list of immutable peptides
        """
        self.input_file = input_file
        self.immutable_peptides = immutable_peptides
        if params is None:
            params = {}
        self.params = params
        self.xml_file_list = self.params.get("xml_file_list", None)
        self.param_mapper = uparma.UParma()
        self.mod_mapper = UnimodMapper(xml_file_list=self.xml_file_list)
        self.params["mapped_mods"] = self.mod_mapper.map_mods(
            mod_list=self.params.get("modifications", [])
        )
        self.mod_dict = self._create_mod_dicts()
        self.cc = ChemicalComposition(
            unimod_file_list=self.params.get("xml_file_list", None)
        )
        self.style = None

    @classmethod
    def check_parser_compatibility(cls, file):
        """Assert compatibility between file and parser.

        Args:
            file (str): path to input file

        Returns:
            bool: True if parser and file are compatible

        """
        return False

    def _load_model(self, file):
        with open(
            Path(__file__).parent / "models" / "base_parser_model.json", "r"
        ) as jf:
            model_definition = json.load(jf)
        with open(Path(__file__).parent / "models" / file, "r") as jf:
            model_definition.extend(json.load(jf))
        model_definition = {item["name"]: item["dtype"] for item in model_definition}
        return model_definition

    def _create_mod_dicts(self):
        """
        Create dict containing meta information about static and variable mods.

        Returns:
            mod_dict (dict): mapped modifications and information
        """
        mod_dict = {}
        for mod_type in ["fix", "opt"]:
            for modification in self.params["mapped_mods"][mod_type]:
                aa = modification["aa"]
                pos = modification["position"]
                name = modification["name"]
                if name not in mod_dict.keys():
                    mod_dict[name] = {
                        "mass": modification["mass"],
                        "aa": set(),
                        "position": set(),
                    }
                mod_dict[name]["aa"].add(aa)

                mod_dict[name]["aa"].add(pos)
                mod_dict[name]["position"].add(pos)

        return mod_dict

    def _read_meta_info_lookup_file(self):
        """Read meta info lookup file.

        Returns:
            rt_lookup (dict of int: dict): dict with spectrum ids as top level key
                                           values are new dicts with all rt values as keys
                                           values are lists with [file, precursor_mz]
        """
        rt_lookup = {}
        with open(self.params["rt_pickle_name"], mode="r") as meta_csv:
            meta_reader = csv.DictReader(meta_csv)
            for row in meta_reader:
                rt = float(row["rt"])
                if row["rt_unit"] == "minute" or row["rt_unit"] == "min":
                    rt *= 60.0
                if int(row["spectrum_id"]) not in rt_lookup:
                    rt_lookup[int(row["spectrum_id"])] = {}
                if row["precursor_mz"] == "":
                    precursor_mz = np.nan
                else:
                    precursor_mz = float(row["precursor_mz"])
                rt_lookup[int(row["spectrum_id"])][rt] = [
                    row["lineage_root"],
                    precursor_mz,
                ]
        return rt_lookup

    def sanitize(self):
        """Perform dataframe sanitation steps.

        - Cast defined types
        - Columns that were not filled in but should exist in the unified format are added and set to None
        - Columns in the dataframe which could not be properly mapped are removed (warning is raised)
        Operations are performed inplace on self.df
        """
        # Set missing columns to None and reorder columns in standardized manner
        col_order = pd.Series(self.required_headers.keys())
        new_cols = col_order[~col_order.isin(self.df.columns)].to_list()
        self.df.loc[:, new_cols] = pd.NA
        self.df = self.df.loc[
            :,
            col_order.tolist()
            + sorted(self.df.columns[~self.df.columns.isin(col_order)].tolist()),
        ]
        str_cols = [c for c, dtype in self.required_headers.items() if dtype == "str"]
        self.df.loc[:, str_cols] = self.df[str_cols].fillna("")
        self.df = self.df.astype(self.required_headers)
        # Ensure there are not any column that should not be
        if hasattr(self, "mapping_dict"):
            new_cols = set(self.mapping_dict.keys())
        else:
            new_cols = set()
        additional_cols = set(self.df.columns).difference(
            set(self.required_headers.keys()) | new_cols
        )
        unmapped_add_cols = [c for c in additional_cols if ":" not in c]
        if len(unmapped_add_cols) > 0:
            logger.warning(
                f"Some engine level columns ({unmapped_add_cols}) were not properly mapped and removed."
            )
            self.df.drop(columns=unmapped_add_cols, inplace=True, errors="ignore")

        # Drop unwanted duplicated rows
        init_len = len(self.df)
        self.df.drop_duplicates(inplace=True)
        rows_dropped = init_len - len(self.df)
        if rows_dropped != 0:
            logger.warning(
                f"{rows_dropped} duplicated rows were dropped in output csv."
            )



#!/usr/bin/env python
import pytest

from pyiohat.parsers.quant.flash_lfq_1_2_0_parser import (
    FlashLFQ_1_2_0_Parser,
)


def test_engine_parsers_flashLFQ_init():
    input_file = pytest._test_path / "data" / "flash_lfq_1_2_0_quantified_peaks.tsv"
    rt_lookup_path = pytest._test_path / "data" / "BSA2_ursgal_lookup.csv"
    parser = FlashLFQ_1_2_0_Parser(
        input_file,
        params={
            "rt_pickle_name": rt_lookup_path,
            "modifications": [
                {
                    "aa": "M",
                    "type": "opt",
                    "position": "any",
                    "name": "Oxidation",
                },
                {
                    "aa": "C",
                    "type": "fix",
                    "position": "any",
                    "name": "Carbamidomethyl",
                },
                {
                    "aa": "*",
                    "type": "opt",
                    "position": "Prot-N-term",
                    "name": "Acetyl",
                },
            ],
        },
    )


def test_engine_parsers_flashLFQ_check_parser_compatibility():
    input_file = pytest._test_path / "data" / "flash_lfq_1_2_0_quantified_peaks.tsv"

    assert FlashLFQ_1_2_0_Parser.check_parser_compatibility(input_file) is True


def test_engine_parsers_flashLFQ_file_not_matches_parser():
    input_file = (
        pytest._test_path / "data" / "test_Creinhardtii_QE_pH11_msfragger_3.tsv"
    )

    assert FlashLFQ_1_2_0_Parser.check_parser_compatibility(input_file) is False


def test_engine_parsers_flashLFQ_unify_row():
    input_file = pytest._test_path / "data" / "flash_lfq_1_2_0_quantified_peaks.tsv"
    rt_lookup_path = pytest._test_path / "data" / "BSA2_ursgal_lookup.csv"

    parser = FlashLFQ_1_2_0_Parser(
        input_file,
        params={
            "rt_pickle_name": rt_lookup_path,
            "modifications": [
                {
                    "aa": "M",
                    "type": "opt",
                    "position": "any",
                    "name": "Oxidation",
                },
                {
                    "aa": "C",
                    "type": "fix",
                    "position": "any",
                    "name": "Carbamidomethyl",
                },
                {
                    "aa": "*",
                    "type": "opt",
                    "position": "Prot-N-term",
                    "name": "Acetyl",
                },
            ],
        },
    )
    df = parser.unify()
    assert len(df) == 11
    assert pytest.approx(df["flashlfq:ms2_retention_time"].mean()) == 118178.01055184663
    assert pytest.approx(df["quant_value"].mean()) == 306668.4375


def test_engine_parsers_flashLFQ_extract_mods():
    input_file = pytest._test_path / "data" / "flash_lfq_1_2_0_quantified_peaks.tsv"
    rt_lookup_path = pytest._test_path / "data" / "BSA2_ursgal_lookup.csv"

    parser = FlashLFQ_1_2_0_Parser(
        input_file,
        params={
            "rt_pickle_name": rt_lookup_path,
            "modifications": [
                {
                    "aa": "M",
                    "type": "opt",
                    "position": "any",
                    "name": "Oxidation",
                },
                {
                    "aa": "C",
                    "type": "fix",
                    "position": "any",
                    "name": "Carbamidomethyl",
                },
                {
                    "aa": "*",
                    "type": "opt",
                    "position": "Prot-N-term",
                    "name": "Acetyl",
                },
            ],
        },
    )
    test_sequence = "ELC[Carbamidomethyl]"
    mods = parser.translate_mods(test_sequence)
    assert mods == "Carbamidomethyl:3"

    test_sequence2 = "ELC[Carbamidomethyl]MMMM[Oxidation]"
    mods = parser.translate_mods(test_sequence2)
    assert mods == "Carbamidomethyl:3;Oxidation:7"

    test_sequence3 = "[Acetyl]ELC[Carbamidomethyl]MMMM[Oxidation]"
    mods = parser.translate_mods(test_sequence3)
    assert mods == "Acetyl:0;Carbamidomethyl:3;Oxidation:7"

    # TODO encode C-terminal mods
    # test_sequence4 = "[Acetyl]ELC[Carbamidomethyl]MMMM[Oxidation][TERMINALMOD]"
    # mods = parser.extract_mods(test_sequence4)
    # assert mods == "Acetyl:0;Carbamidomethyl:3;Oxidation:7;TERMINALMOD:8"



"""Engine parser."""
import itertools

import pandas as pd
import regex as re
from loguru import logger

from pyiohat.parsers.ident_base_parser import IdentBaseParser


class MSFragger_3_Parser(IdentBaseParser):
    """File parser for MSFragger 3."""

    def __init__(self, *args, **kwargs):
        """Initialize parser.

        Reads in data file and provides mappings.
        """
        super().__init__(*args, **kwargs)
        self.style = "msfragger_style_3"
        # 15N handling missing for now
        if self.params.get("label", "") == "15N":
            raise NotImplementedError

        self.df = pd.read_csv(self.input_file, delimiter="\t")
        self.df.dropna(axis=1, how="all", inplace=True)

        self.mapping_dict = {
            v: k
            for k, v in self.param_mapper.get_default_params(style=self.style)[
                "header_translations"
            ]["translated_value"].items()
        }
        self.df.rename(columns=self.mapping_dict, inplace=True)
        self.df.columns = self.df.columns.str.lstrip(" ")
        if not "modifications" in self.df.columns:
            self.df["modifications"] = ""
        self.reference_dict.update({k: None for k in self.mapping_dict.values()})

    @classmethod
    def check_parser_compatibility(cls, file):
        """Assert compatibility between file and parser.

        Args:
            file (str): path to input file

        Returns:
            bool: True if parser and file are compatible

        """
        is_tsv = file.as_posix().endswith(".tsv")
        with open(file.as_posix()) as f:
            try:
                head = "".join([next(f) for _ in range(1)])
            except StopIteration:
                head = ""
        head = set(head.rstrip("\n").split("\t"))
        ref_columns = {
            "scannum",
            "peptide",
            "charge",
            "peptide_prev_aa",
            "peptide_next_aa",
            "protein",
            "modification_info",
            "retention_time",
            "precursor_neutral_mass",
            "calc_neutral_pep_mass",
            "hit_rank",
            "massdiff",
            "num_matched_ions",
            "tot_num_ions",
            "hyperscore",
            "nextscore",
            "num_tol_term",
            "num_missed_cleavages",
            "expectscore",
            "best_locs",
            "score_without_delta_mass",
            "best_score_with_delta_mass",
            "second_best_score_with_delta_mass",
            "delta_score",
        }
        columns_match = len(ref_columns.difference(head)) == 0
        return is_tsv and columns_match

    def _map_mod_translation(self, row, map_dict):
        """Replace single mod string.

        Args:
            row (str): unprocessed modification string
            map_dict (dict): mod mapping dict

        Returns:
            mod_str (str): formatted modification string
        """
        mod_str = ""
        if row == "" or row == [""]:
            return mod_str
        for mod in row:
            mass = re.search(r"\(([^)]+)", mod).group(1)
            name = map_dict[mass]
            if len(name) > 0:
                for m in name:
                    pos = int(re.search(r"^\d+", mod).group(0))
                    if (
                        any(["N-term" in p for p in self.mod_dict[m]["position"]])
                        and pos == 1
                    ):
                        pos = 0
                    else:
                        pos = int(re.search(r"^\d+", mod).group(0))
                    mod_str += f"{m}:{pos};"
            else:
                return "NON_MAPPABLE"
        return mod_str

    def translate_mods(self):
        """
        Replace internal modification nomenclature with formatted modification strings.

        Returns:
            (pd.Series): column with formatted mod strings
        """
        mod_split_col = self.df["modifications"].fillna("").str.split(", ")
        unique_mods = set().union(*mod_split_col.apply(set)).difference({""})
        unique_mod_masses = {re.search(r"\(([^)]+)", m).group(1) for m in unique_mods}
        # Map single mods
        potential_names = {
            m: [
                name
                for name in self.mod_mapper.mass_to_names(float(m), decimals=4)
                if name in self.mod_dict
            ]
            for m in unique_mod_masses
        }
        # Map multiple mods
        for n in [2, 3]:
            for unmapped_mass in {k: v for k, v in potential_names.items() if v == []}:
                potential_mods = [
                    name[1]
                    for name in self.mod_mapper.mass_to_combos(
                        float(unmapped_mass), n=n, decimals=4
                    )
                    if all(m in self.mod_dict for m in name[1])
                ]
                if len(potential_mods) == 1:
                    potential_names[unmapped_mass] = potential_mods[0]
        non_mappable_mods = {
            k: len(
                [
                    m
                    for m in list(
                        itertools.chain.from_iterable(
                            mod_split_col.apply(list).to_list()
                        )
                    )
                    if k in m
                ]
            )
            for k, v in potential_names.items()
            if v == []
        }
        non_mappable_percent = pd.Series(
            [v / len(self.df) for v in non_mappable_mods.values()], dtype="float64"
        )
        if any(non_mappable_percent > 0.001):
            raise ValueError(
                "Some modifications found in more than 0.1% of PSMs cannot be mapped."
            )
        if len(non_mappable_percent) > 0:
            logger.warning(
                "Some modifications found in less than 0.1% of PSMs cannot be mapped and were removed."
            )
        mods_translated = mod_split_col.apply(
            self._map_mod_translation, map_dict=potential_names
        )

        return mods_translated.str.rstrip(";")

    def unify(self):
        """
        Primary method to read and unify engine output.

        Returns:
            self.df (pd.DataFrame): unified dataframe
        """
        self.df["search_engine"] = "msfragger_3_0"
        self.df["retention_time_seconds"] *= 60.0
        self.df["exp_mz"] = self._calc_mz(
            mass=self.df["msfragger:precursor_neutral_mass_da"],
            charge=self.df["charge"],
        )
        self.df["modifications"] = self.translate_mods()
        self.df = self.df.loc[
            ~self.df["modifications"].str.contains("NON_MAPPABLE", regex=False), :
        ]
        self.process_unify_style()

        return self.df



from pathlib import Path

import numpy as np
import pandas as pd
import pytest

from pyiohat.parsers.ident_base_parser import IdentBaseParser
from pyiohat.utils import merge_and_join_dicts


def test_base_parser_read_rt_lookup_file():
    rt_lookup_path = pytest._test_path / "data" / "BSA1_ursgal_lookup.csv"
    input_file = (
        pytest._test_path / "data" / "test_Creinhardtii_QE_pH11_xtandem_alanine.xml"
    )

    bp = IdentBaseParser(input_file, params={"rt_pickle_name": rt_lookup_path})
    rt_lookup = bp._read_meta_info_lookup_file()
    assert len(rt_lookup) == 1120
    precursor_mzs = [list(specs.values())[-1][-1] for specs in rt_lookup.values()]
    assert pytest.approx(sum(precursor_mzs) / len(precursor_mzs)) == 550.8444810049874
    # check consistency
    assert 2450 in rt_lookup
    assert 1534.4619140625 in rt_lookup[2450]
    assert rt_lookup[2450][1534.4619140625][0] == "path/for/glory.mzML"
    assert pytest.approx(rt_lookup[2450][1534.4619140625][1]) == 618.2697


def test_engine_parsers_IdentBaseParser_init():
    input_file = (
        pytest._test_path / "data" / "test_Creinhardtii_QE_pH11_xtandem_alanine.xml"
    )
    rt_lookup_path = pytest._test_path / "data" / "_ursgal_lookup.csv"
    db_path = pytest._test_path / "data" / "test_Creinhardtii_target_decoy.fasta"

    parser = IdentBaseParser(
        input_file,
        params={
            "cpus": 2,
            "rt_pickle_name": rt_lookup_path,
            "database": db_path,
            "modifications": [
                {
                    "aa": "M",
                    "type": "opt",
                    "position": "any",
                    "name": "Oxidation",
                },
                {
                    "aa": "C",
                    "type": "fix",
                    "position": "any",
                    "name": "Carbamidomethyl",
                },
                {
                    "aa": "*",
                    "type": "opt",
                    "position": "Prot-N-term",
                    "name": "Acetyl",
                },
            ],
        },
    )


def test_engine_parsers_IdentBaseParser_check_parser_compatibility_non_existing():
    # should always return False
    IdentBaseParser.check_parser_compatibility("whatever") is False


def test_engine_parsers_IdentBaseParser_check_parser_compatibility_existing():
    # should always return False
    IdentBaseParser.check_parser_compatibility(
        pytest._test_path / "data" / "test_Creinhardtii_QE_pH11_xtandem_alanine.xml"
    ) is False


def test_engine_parsers_IdentBase_Parser_sanitize():
    obj = IdentBaseParser(input_file=None, params=None)
    obj.mapping_dict = {"Engine:C": None, "Engine:B": None, "Engine:A": None}
    obj.df = pd.DataFrame(
        np.random.random((5, len(obj.col_order) + 4)),
        columns=obj.col_order.to_list()
        + ["Engine:C", "Engine:B", "Engine:A", "This should not exist"],
    )
    str_cols = [c for c, dtype in obj.required_headers.items() if dtype == "str"]
    int_cols = [c for c, dtype in obj.required_headers.items() if dtype == "Int32"]
    bool_cols = [c for c, dtype in obj.required_headers.items() if dtype == "bool"]
    obj.df.loc[:, str_cols] = "4"
    obj.df.loc[:, int_cols] = "2"
    obj.df.loc[:, bool_cols] = False
    obj.df["raw_data_location"] = "test.mgf"
    obj.df["spectrum_title"] = "spec_title.this_should_go"
    obj.df.loc[3, "raw_data_location"] = np.nan
    obj.df["modifications"] = "ZZ:1;AA:8"
    obj.sanitize()
    assert obj.df.loc[3, "raw_data_location"] == ""
    assert (obj.df["modifications"] == "ZZ:1;AA:8").all()
    assert obj.df.columns.to_list() == obj.col_order.to_list() + [
        "Engine:A",
        "Engine:B",
        "Engine:C",
    ]


def test_add_ranks_increasing_engine_scores_better():
    obj = IdentBaseParser(
        input_file=None,
        params={
            "validation_score_field": {"msfragger_3_0": "msfragger:hyperscore"},
            "bigger_scores_better": {"msfragger_3_0": True},
        },
    )
    obj.df = pd.DataFrame(
        np.ones((5, len(obj.col_order) + 1)),
        columns=obj.col_order.to_list() + ["msfragger:hyperscore"],
    )
    obj.df["msfragger:hyperscore"] = [5, 2, 1, 3, 3]
    obj.df["search_engine"] = "msfragger_3_0"
    obj.add_ranks()
    assert obj.df["rank"].to_list() == [1, 4, 5, 2, 2]


def test_add_ranks_decreasing_engine_scores_better():
    obj = IdentBaseParser(
        input_file=None,
        params={
            "validation_score_field": {"msgfplus_2021_03_22": "ms-gf:spec_evalue"},
            "bigger_scores_better": {"msgfplus_2021_03_22": False},
        },
    )
    obj.df = pd.DataFrame(
        np.ones((5, len(obj.col_order) + 1)),
        columns=obj.col_order.to_list() + ["msfragger:hyperscore"],
    )
    obj.df["ms-gf:spec_evalue"] = [5, 2, 1, 3, 3]
    obj.df["search_engine"] = "msgfplus_2021_03_22"
    obj.add_ranks()
    assert obj.df["rank"].to_list() == [5, 2, 1, 3, 3]


def test_add_protein_ids():
    obj = IdentBaseParser(
        input_file=None,
        params={
            "cpus": 2,
            "database": pytest._test_path / "data/test_Creinhardtii_target_decoy.fasta",
            "validation_score_field": {"msfragger_3_0": "msfragger:hyperscore"},
            "bigger_scores_better": {"msfragger_3_0": True},
        },
    )
    obj.df = pd.DataFrame(
        np.ones((1, 1)),
        columns=["msfragger:hyperscore"],
    )
    obj.df.loc[0, "sequence"] = "SAVVGTFFR"
    obj.add_protein_ids()

    assert set(obj.df.columns) == {
        "sequence_stop",
        "sequence",
        "protein_id",
        "sequence_post_aa",
        "sequence_pre_aa",
        "sequence_start",
        "msfragger:hyperscore",
    }
    assert obj.df.loc[0, "sequence_start"] == "287"
    assert obj.df.loc[0, "sequence_stop"] == "295"
    assert obj.df.loc[0, "sequence_pre_aa"] == "R"
    assert obj.df.loc[0, "sequence_post_aa"] == "D"
    assert (
        obj.df.loc[0, "protein_id"]
        == "Cre12.g514050.t1.2 pacid=30792611 transcript=Cre12.g514050.t1.2 locus=Cre12.g514050 ID=Cre12.g514050.t1.2.v5.5 annot-version=v5.5"
    )


def test_calc_masses_offsets_and_composition():
    obj = IdentBaseParser(
        input_file=None,
        params={
            "cpus": 2,
            "rt_pickle_name": pytest._test_path / "data/_ursgal_lookup.csv",
            "validation_score_field": {"msfragger_3_0": "msfragger:hyperscore"},
            "bigger_scores_better": {"msfragger_3_0": True},
            "modifications": [
                {
                    "aa": "*",
                    "type": "opt",
                    "position": "Prot-N-term",
                    "name": "Acetyl",
                },
                {
                    "aa": "C",
                    "type": "fix",
                    "position": "any",
                    "name": "Carbamidomethyl",
                },
            ],
        },
    )
    obj.df = pd.DataFrame(
        np.ones((3, len(obj.col_order) + 1)),
        columns=obj.col_order.to_list() + ["msfragger:hyperscore"],
    )
    obj.df["spectrum_id"] = 3 * [10152]
    obj.df["retention_time_seconds"] = 3 * [1912.25016]
    obj.get_meta_info()
    obj.df.loc[:, "sequence"] = 3 * ["PEPTCIDE"]
    obj.df.loc[:, "modifications"] = [
        "",
        "Carbamidomethyl:5",
        "Acetyl:0;Carbamidomethyl:5",
    ]
    obj.calc_masses_offsets_and_composition()
    ref_masses = np.array(
        [902.369149, 902.369149 + 57.021464, 902.369149 + 57.021464 + 42.010565]
    )

    assert np.allclose(obj.df["ucalc_mass"], ref_masses, atol=1e-4)
    assert np.allclose(obj.df["ucalc_mz"], ref_masses + obj.PROTON, atol=1e-4)
    assert np.allclose(
        obj.df["accuracy_ppm"], [-159398.0953, -209306.9420, -242444.5934], atol=1e-3
    )
    assert np.allclose(
        obj.df["accuracy_ppm_C12"],
        [-159398.0953, -209306.9420, -242444.5933],
        atol=1e-3,
    )


def test_get_exp_rt_and_mz():
    obj = IdentBaseParser(
        input_file=None,
        params={
            "cpus": 2,
            "rt_pickle_name": pytest._test_path / "data/_ursgal_lookup.csv",
        },
    )
    obj.df = pd.DataFrame(
        np.ones((5, len(obj.col_order) + 1)),
        columns=obj.col_order.to_list() + ["msfragger:hyperscore"],
    )
    obj.df["spectrum_id"] = [10152, 10381, 10414, 10581, 11535]
    obj.df["retention_time_seconds"] = [
        1912.25016,
        1943.05878,
        1946.75502,
        1967.22444,
        2082.79998,
    ]
    obj.get_meta_info()
    assert np.allclose(
        obj.df["exp_mz"], [759.379, 439.196, 739.358, 664.286, 496.264], atol=1e-3
    )
    assert np.allclose(
        obj.df["retention_time_seconds"],
        [1912.25, 1943.0586666666668, 1946.755, 1967.2243333333333, 2082.7998333333335],
        atol=1e-2,
    )


def test_create_mod_dicts():
    obj = IdentBaseParser(
        input_file=None,
        params={
            "cpus": 2,
            "modifications": [
                {
                    "aa": "*",
                    "type": "opt",
                    "position": "Prot-N-term",
                    "name": "Acetyl",
                },
                {
                    "aa": "C",
                    "type": "fix",
                    "position": "any",
                    "name": "Carbamidomethyl",
                },
            ],
        },
    )
    mod_dict = obj._create_mod_dicts()
    reference_dict = {
        "Carbamidomethyl": {"mass": 57.021464, "aa": {"C", "any"}, "position": {"any"}},
        "Acetyl": {
            "mass": 42.010565,
            "aa": {"Prot-N-term", "*"},
            "position": {"Prot-N-term"},
        },
    }
    assert mod_dict == reference_dict


def test_calc_mz():
    obj = IdentBaseParser(input_file=None, params=None)
    masses_in_weird_types = pd.Series(["10", 20, 30.0])
    charges_in_weird_types = pd.Series(["2", 2, 2.0])
    mz = obj._calc_mz(masses_in_weird_types, charges_in_weird_types).to_list()

    assert mz[0] == pytest.approx(6.0, abs=1e-2)
    assert mz[1] == pytest.approx(11.0, abs=1e-2)
    assert mz[2] == pytest.approx(16.0, abs=1e-2)


def test_calc_mass():
    obj = IdentBaseParser(input_file=None, params=None)
    mzs_in_weird_types = pd.Series(["6", 11, 16.0])
    charges_in_weird_types = pd.Series(["2", 2, 2.0])
    mz = obj._calc_mass(mzs_in_weird_types, charges_in_weird_types).to_list()

    assert mz[0] == pytest.approx(10.0, abs=1e-1)
    assert mz[1] == pytest.approx(20.0, abs=1e-1)
    assert mz[2] == pytest.approx(30.0, abs=1e-1)


def test_get_closest_isotopologue():
    obj = IdentBaseParser(
        input_file=None,
        params={
            "cpus": 2,
            "modifications": [
                {
                    "aa": "*",
                    "type": "opt",
                    "position": "Prot-N-term",
                    "name": "Acetyl",
                },
                {
                    "aa": "C",
                    "type": "fix",
                    "position": "any",
                    "name": "Carbamidomethyl",
                },
            ],
        },
    )
    obj.df = pd.DataFrame(
        np.ones((2, len(obj.col_order) + 1)),
        columns=obj.col_order.to_list() + ["msfragger:hyperscore"],
    )
    obj.df["sequence"] = ["PEPTIDE", "ASDF"]
    obj.df["modifications"] = ["Acetyl:2", ""]
    obj.df["charge"] = [1, 1]
    obj.df["exp_mz"] = [840.0, 441.0]
    obj.calc_masses_offsets_and_composition()
    assert np.allclose(
        obj.df.loc[:, "accuracy_ppm"], [-2822.7301, -428.5010], atol=1e-3
    )
    assert np.allclose(
        obj.df.loc[:, "accuracy_ppm_C12"], [-2822.7300, 4138.7359], atol=1e-3
    )
    obj.df["exp_mz"] = [845.0, 420.0]
    obj.calc_masses_offsets_and_composition()
    assert np.allclose(
        obj.df.loc[:, "accuracy_ppm"], [728.9157, -43677.3943], atol=1e-3
    )
    assert np.allclose(
        obj.df.loc[:, "accuracy_ppm_C12"], [3112.8488, -43677.3943], atol=1e-3
    )


def test_get_composition():
    obj = IdentBaseParser(
        input_file=None,
        params={
            "cpus": 2,
            "modifications": [
                {
                    "aa": "*",
                    "type": "opt",
                    "position": "Prot-N-term",
                    "name": "Acetyl",
                },
                {
                    "aa": "C",
                    "type": "fix",
                    "position": "any",
                    "name": "Carbamidomethyl",
                },
                {
                    "aa": "*",
                    "type": "opt",
                    "position": "any",
                    "name": "Label:18O(1)",
                },
                {
                    "aa": "*",
                    "type": "opt",
                    "position": "any",
                    "name": "CustomMod42",
                },
            ],
            "xml_file_list": [Path(pytest._test_path / "data" / "custom_mod.xml")],
        },
    )
    obj.df = pd.DataFrame(
        np.ones((4, len(obj.col_order) + 1)),
        columns=obj.col_order.to_list() + ["msfragger:hyperscore"],
    )
    obj.df["sequence"] = 4 * ["PEPTCIDE"]
    obj.df["modifications"] = [
        "",
        "Acetyl:0;Carbamidomethyl:5",
        "Acetyl:0;Carbamidomethyl:5;Label:18O(1):7",
        "CustomMod42:4",
    ]
    seq = "PEPTCIDE"
    obj.calc_masses_offsets_and_composition()

    # Without modifications
    assert obj.df.loc[0, "chemical_composition"] == "C(37)H(58)N(8)O(16)S(1)"

    # With modifications
    assert obj.df.loc[1, "chemical_composition"] == "C(41)H(63)N(9)O(18)S(1)"

    # With labeled modifications
    assert obj.df.loc[2, "chemical_composition"] == "C(41)H(63)18O(1)N(9)O(17)S(1)"

    # Change the ChemicalComposition instance to use custom mods
    assert obj.df.loc[3, "chemical_composition"] == "C(37)H(61)13C(3)N(8)O(16)S(1)"


def test_merge_and_join_dicts():
    dict_a = {"a": "part_a", "b": "part_a"}
    dict_b = {"a": "part_b", "b": "part_b"}
    out_dict = merge_and_join_dicts([dict_a, dict_b], ";")
    assert out_dict == {"a": "part_a;part_b", "b": "part_a;part_b"}


def test_assert_only_iupac_and_missing_aas():
    obj = IdentBaseParser(
        input_file=None,
        params={
            "cpus": 2,
            "rt_pickle_name": pytest._test_path / "data/_ursgal_lookup.csv",
        },
    )
    obj.df = pd.DataFrame(
        np.ones((4, len(obj.col_order) + 1)),
        columns=obj.col_order.to_list() + ["msfragger:hyperscore"],
    )
    obj.df["sequence"] = ["PEPTIDE", "[3jd2]", "ACDXU", "MREPEPTIDE"]
    obj.assert_only_iupac_and_missing_aas()

    assert all(obj.df.index == [0, 3])


def test_add_decoy_identity():
    obj = IdentBaseParser(
        input_file=None,
        params={
            "cpus": 2,
            "rt_pickle_name": pytest._test_path / "data/_ursgal_lookup.csv",
        },
    )
    obj.df = pd.DataFrame(
        np.ones((4, len(obj.col_order) + 1)),
        columns=obj.col_order.to_list() + ["msfragger:hyperscore"],
    )
    obj.df["protein_id"] = ["NOTADECOY", "PEPTIDE", "decoy_PEPTIDE", "decoy_ASDF"]

    obj.add_decoy_identity()

    assert all(obj.df["is_decoy"] == [False, False, True, True])


def test_add_decoy_identity_non_default_prefix():
    obj = IdentBaseParser(
        input_file=None,
        params={
            "cpus": 2,
            "rt_pickle_name": pytest._test_path / "data/_ursgal_lookup.csv",
            "decoy_tag": "non_default_tag_",
        },
    )
    obj.df = pd.DataFrame(
        np.ones((4, len(obj.col_order) + 1)),
        columns=obj.col_order.to_list() + ["msfragger:hyperscore"],
    )
    obj.df["protein_id"] = [
        "NOTADECOY",
        "PEPTIDE",
        "decoy_PEPTIDE",
        "non_default_tag_ASDF",
    ]

    obj.add_decoy_identity()

    assert all(obj.df["is_decoy"] == [False, False, False, True])


def test_add_decoy_identity_with_immutable_peptides():
    obj = IdentBaseParser(
        input_file=None,
        params={
            "cpus": 2,
            "rt_pickle_name": pytest._test_path / "data/_ursgal_lookup.csv",
        },
        immutable_peptides=["GONEIN", "UPAND"],
    )
    obj.df = pd.DataFrame(
        np.ones((5, len(obj.col_order) + 1)),
        columns=obj.col_order.to_list() + ["msfragger:hyperscore"],
    )
    obj.df["protein_id"] = [
        "NOTADECOY",
        "PEPTIDE",
        "decoy_PEPTIDE",
        "decoy_ASDF",
        "decoy_BUTIMMUTABLE",
    ]
    obj.df["sequence"] = [
        "GONEINTHEWIND",
        "GONEIN",
        "UPAND",
        "UPANDAWAY",
        "GONEINUPAND",
    ]

    obj.add_decoy_identity()
    assert all(obj.df["is_immutable"] == [False, True, True, False, True])


def test_check_enzyme_specificity_trypsin_all():
    obj = IdentBaseParser(
        input_file=None,
        params={
            "cpus": 2,
            "rt_pickle_name": pytest._test_path / "data/_ursgal_lookup.csv",
            "enzyme": "(?<=[KR])(?![P])",
            "terminal_cleavage_site_integrity": "all",
        },
    )
    obj.df = pd.DataFrame(
        np.ones((4, len(obj.col_order) + 1)),
        columns=obj.col_order.to_list() + ["msfragger:hyperscore"],
    )
    obj.df["sequence"] = ["PEPRTIDEK", "EPTIDEK", "EPTIDEK", "EPRPTRIRDEK"]
    obj.df["sequence_pre_aa"] = ["K", "K", "A", "K"]
    obj.df["sequence_post_aa"] = ["A", "P", "A<|>P", "A<|>V"]
    obj.check_enzyme_specificity()

    assert all(obj.df["enzn"] == [False, True, False, True])
    assert all(obj.df["enzc"] == [True, False, False, True])
    assert all(obj.df["missed_cleavages"] == [1, 0, 0, 2])


def test_check_enzyme_specificity_trypsin_any():
    obj = IdentBaseParser(
        input_file=None,
        params={
            "cpus": 2,
            "rt_pickle_name": pytest._test_path / "data/_ursgal_lookup.csv",
            "enzyme": "(?<=[KR])(?![P])",
            "terminal_cleavage_site_integrity": "any",
        },
    )
    obj.df = pd.DataFrame(
        np.ones((4, len(obj.col_order) + 1)),
        columns=obj.col_order.to_list() + ["msfragger:hyperscore"],
    )
    obj.df["sequence"] = ["PEPRTIDEK", "EPTIDEK", "EPTIDEK", "EPTIDEK"]
    obj.df["sequence_pre_aa"] = ["K", "K", "A", "K"]
    obj.df["sequence_post_aa"] = ["A", "P", "A<|>P", "A<|>V"]
    obj.check_enzyme_specificity()

    assert all(obj.df["enzn"] == [False, True, False, True])
    assert all(obj.df["enzc"] == [True, False, True, True])
    assert all(obj.df["missed_cleavages"] == [1, 0, 0, 0])


def test_check_enzyme_specificity_nonspecific():
    obj = IdentBaseParser(
        input_file=None,
        params={
            "cpus": 2,
            "rt_pickle_name": pytest._test_path / "data/_ursgal_lookup.csv",
            "enzyme": ".^",
            "terminal_cleavage_site_integrity": "all",
        },
    )
    obj.df = pd.DataFrame(
        np.ones((4, len(obj.col_order) + 1)),
        columns=obj.col_order.to_list() + ["msfragger:hyperscore"],
    )
    obj.df["sequence"] = ["ASDF", "EPTIDEK", "EPTIDEK", "EPTIDEK"]
    obj.df["sequence_pre_aa"] = ["L<|>E", "A", "R", "P"]
    obj.df["sequence_post_aa"] = ["F", "L<|>E", "A<|>P", "A<|>V"]
    obj.check_enzyme_specificity()

    assert all(obj.df["enzn"] == [True, True, True, True])
    assert all(obj.df["enzc"] == [True, True, True, True])
    assert all(obj.df["missed_cleavages"] == [0, 0, 0, 0])


def test_groupby_rt_and_spec_id():
    obj = IdentBaseParser(
        input_file=None,
        params={
            "cpus": 2,
            "rt_pickle_name": pytest._test_path / "data/_ursgal_lookup.csv",
        },
    )
    obj.df = pd.DataFrame(
        np.ones((3, len(obj.col_order) + 1)),
        columns=obj.col_order.to_list() + ["msfragger:hyperscore"],
    )
    obj.df["spectrum_id"] = [10152, 10381, 10414]
    obj.df["retention_time_seconds"] = [
        1912.25016,  # Identical
        1943.058,  # Very similar
        1946.76,  # Similar
    ]
    obj.get_meta_info()
    assert (
        all(obj.df["exp_mz"] == [759.379943847656, 439.196746826172, 739.358459472656])
        is True
    )
    assert (
        all(obj.df["retention_time_seconds"] == [1912.25016, 1943.05878, 1946.75502])
        is True
    )
    with pytest.raises(KeyError):
        obj.df.loc[0, "spectrum_id"] = 1234567
        obj.get_meta_info()


def test_clean_up_modifications():
    obj = IdentBaseParser(
        input_file=None,
        params={
            "cpus": 2,
            "rt_pickle_name": pytest._test_path / "data/_ursgal_lookup.csv",
            "enzyme": "(?<=[KR])(?![P])",
            "terminal_cleavage_site_integrity": "any",
        },
    )
    obj.df = pd.DataFrame()
    obj.df["modifications"] = [
        "TMTpro:12;;Oxidation:2",
        "TMTpro:10;Acetyl:0",
        "",
        "Acetyl:0;Oxidation:5",
        "Label:18O(1):7;Acetyl:0;Oxidation:5",
    ]
    expected_mods = [
        "Oxidation:2;TMTpro:12",
        "Acetyl:0;TMTpro:10",
        "",
        "Acetyl:0;Oxidation:5",
        "Acetyl:0;Oxidation:5;Label:18O(1):7",
    ]

    obj.clean_up_modifications()
    print(obj.df["modifications"])
    assert (obj.df["modifications"] == expected_mods).all()



import numpy as np

from pyiohat.parsers.misc import get_atom_counts


def test_simple():
    compositions = {  # these include water ...
        "E": {"C": 5, "H": 9, "N": 1, "O": 4},
        "Magic": {"H": 100},
    }
    modifications = np.array(
        [
            "",
            "",
            "Magic:1",
        ],
        dtype=str,
    )
    sequences = np.array(
        [
            "E",
            "EE",
            "E",
        ],
        dtype=str,
    )
    elements, atom_counts = get_atom_counts(
        sequences=sequences,
        modifications=modifications,
        compositions=compositions,
    )

    assert np.array_equal(elements, ["C", "H", "N", "O"])
    assert np.array_equal(atom_counts[0], [5, 9, 1, 4])
    assert np.array_equal(atom_counts[1], [10, 16, 2, 7])
    assert np.array_equal(atom_counts[2], [5, 109, 1, 4])


def test_negative():
    compositions = {  # these include water ...
        "E": {"C": 5, "H": 9, "N": 1, "O": 4},
        "Magic": {"H": -10},
    }
    modifications = np.array(
        [
            "Magic:1",
        ],
        dtype=str,
    )
    sequences = np.array(
        [
            "E",
        ],
        dtype=str,
    )
    elements, atom_counts = get_atom_counts(
        sequences=sequences,
        modifications=modifications,
        compositions=compositions,
    )
    assert np.array_equal(atom_counts[0], [5, -1, 1, 4])



import numpy as np
import pytest

from pyiohat.parsers.base_parser import BaseParser


def test_uninitialized_parser_compatiblity_is_false():
    input_file = (
        pytest._test_path / "data" / "test_Creinhardtii_QE_pH11_xtandem_alanine.xml"
    )
    compat = BaseParser.check_parser_compatibility(input_file)
    assert compat is False


def test_base_parser_read_rt_lookup_file_wo_precursor_mz_info():
    rt_lookup_path = (
        pytest._test_path / "data" / "BSA1_ursgal_lookup_no_precursor_mz.csv"
    )
    input_file = (
        pytest._test_path / "data" / "test_Creinhardtii_QE_pH11_xtandem_alanine.xml"
    )

    bp = BaseParser(input_file, params={"rt_pickle_name": rt_lookup_path})
    rt_lookup = bp._read_meta_info_lookup_file()
    assert len(rt_lookup) == 9
    precursor_mzs = [list(specs.values())[-1][-1] for specs in rt_lookup.values()]
    assert set(precursor_mzs) == {np.nan}
    # check consistency
    assert 2450 in rt_lookup
    assert 1534.4619140625 in rt_lookup[2450]
    assert rt_lookup[2450][1534.4619140625][0] == "path/for/glory.mzML"
    assert rt_lookup[2450][1534.4619140625][1] is np.nan



"""Ident base parser class."""
import csv
import multiprocessing as mp

import ahocorasick
import numpy as np
import pandas as pd
from chemical_composition.chemical_composition_kb import PROTON
from loguru import logger
from peptide_mapper.mapper import UPeptideMapper


from pyiohat.parsers.base_parser import BaseParser
from pyiohat.parsers.misc import (
    get_compositions_and_monoisotopic_masses,
    get_isotopologue_accuracy,
    init_custom_cc,
    sort_mods,
)
from pyiohat.utils import merge_and_join_dicts

mp.set_start_method(method="fork")


class IdentBaseParser(BaseParser):
    """Base class of all ident parsers."""

    def __init__(self, *args, **kwargs):
        """Initialize parser.

        Reads in data file and provides mappings.
        """
        super().__init__(*args, **kwargs)
        self.DELIMITER = self.params.get("delimiter", "<|>")
        self.PROTON = PROTON
        self.IUPAC_AAS = tuple("ACDEFGHIKLMNPQRSTUVWY")
        self.df = None

        successfully_mapped_mods = set(
            mod["name"] for mod in self.params["mapped_mods"]["fix"]
        ) | set(mod["name"] for mod in self.params["mapped_mods"]["opt"])
        self.non_mappable_mods = set(
            mod["name"] for mod in self.params.get("modifications", [])
        ).difference(successfully_mapped_mods)
        self.mod_dict = self._create_mod_dicts()
        self.rt_truncate_precision = 2
        self.reference_dict = {
            "search_engine": None,
            "spectrum_id": None,
            "modifications": None,
            "retention_time_seconds": None,
        }
        self.required_headers = self._load_model("ident_parser_model.json")
        self.col_order = pd.Series(self.required_headers.keys())

    def _calc_mz(self, mass, charge):
        """Calculate mass-to-charge ratio.

        Args:
            mass (pd.Series): masses
            charge (pd.Series): charges
        Returns:
            (pd.Series): m/z
        """
        return (
            mass.astype(float) + (charge.astype(int) * self.PROTON)
        ) / charge.astype(int)

    def _calc_mass(self, mz, charge):
        """Calculate mass from mass-to-charge ratio.

        Args:
            mz (pd.Series): mass-to-charge
            charge (pd.Series): charges
        Returns:
            (pd.Series): mass
        """
        return mz.astype(float) * charge.astype(int) - (
            charge.astype(int) * self.PROTON
        )

    def clean_up_modifications(self):
        """Sanitizes modstrings generated by engine parsers.

        Modifications are sorted by position and leading, repeated or trailing delimiters are removed
        Operations are performed inplace on self.df
        """
        # Remove any trailing, leading, repeated, or only delimiters modstrings
        self.df.loc[:, "modifications"] = self.df.loc[:, "modifications"].str.replace(
            r"^;+(?=\w)|(?<=\w);+$|^;+$|;+(?=;)", "", regex=True
        )

        # Ensure same order of modifications
        self.df.loc[:, "modifications"] = (
            self.df["modifications"].fillna("").str.split(";").apply(sort_mods)
        )

    def assert_only_iupac_and_missing_aas(self):
        """Assert that only IUPAC nomenclature one letter amino acids are used in sequence.

        Non-IUPAC designations are dropped (except for selenocysteine).
        Operations are performed inplace.
        """
        self.df["sequence"] = self.df["sequence"].str.upper()
        # Added X for missing AAs
        iupac_conform_seqs = self.df["sequence"].apply(
            lambda seq: set(seq).issubset(self.IUPAC_AAS)
        )
        if any(~iupac_conform_seqs):
            self.df = self.df.loc[iupac_conform_seqs, :]
            logger.warning(
                f"sequences are not IUPAC conform. {(~iupac_conform_seqs).sum()} PSMs were dropped."
            )

    def add_protein_ids(self):
        """Add all Protein IDs that matching the sequence.

        Operations are performed inplace on self.df
        """
        peptide_mapper = UPeptideMapper(self.params["database"])
        mapped_peptides = peptide_mapper.map_peptides(self.df["sequence"].tolist())

        peptide_mappings = [
            merge_and_join_dicts(mapped_peptides[seq], self.DELIMITER)
            for seq in self.df["sequence"]
        ]

        columns_translations = {
            "start": "sequence_start",
            "end": "sequence_stop",
            "post": "sequence_post_aa",
            "id": "protein_id",
            "pre": "sequence_pre_aa",
        }
        new_columns = pd.DataFrame(peptide_mappings)
        new_columns.rename(columns=columns_translations, inplace=True)

        self.df.loc[:, new_columns.columns] = new_columns.values
        new_columns = new_columns.dropna(axis=0, how="all")
        if len(new_columns) != len(self.df):
            logger.warning(
                f"{len(self.df)-len(new_columns)} PSMs were dropped because their respective sequences could not be mapped."
            )
        self.df = self.df.iloc[new_columns.index, :].reset_index(drop=True)

    def check_enzyme_specificity(self):
        """Check consistency of N/C-terminal cleavage sites.

        Calculates number of missed cleavage sites.
        Operations are performed inplace.
        """
        if self.params["enzyme"] == ".^":
            self.df.loc[:, ["enzn", "enzc"]] = True
            self.df.loc[:, "missed_cleavages"] = 0
            return None

        enzyme_pattern = self.params["enzyme"]
        integrity_strictness = self.params["terminal_cleavage_site_integrity"]

        pren_seq = (
            pd.concat(
                [
                    self.df["sequence_pre_aa"].str.split(rf"{self.DELIMITER}"),
                    self.df["sequence"].str[:1],
                ],
                axis=1,
            )
            .explode("sequence_pre_aa")
            .sum(axis=1)
        )
        self.df.loc[:, "enzn"] = (
            pren_seq.str.split(rf"{enzyme_pattern}").str[0].str.len() == 1
        ).groupby(pren_seq.index).agg(integrity_strictness) | (
            pren_seq.str[0] == "-"
        ).groupby(
            pren_seq.index
        ).agg(
            integrity_strictness
        )
        postc_seq = (
            pd.concat(
                [
                    self.df["sequence"].str[-1:],
                    self.df["sequence_post_aa"].str.split("<\\|>"),
                ],
                axis=1,
            )
            .explode("sequence_post_aa")
            .sum(axis=1)
        )
        self.df.loc[:, "enzc"] = (
            postc_seq.str.split(rf"{enzyme_pattern}").str[0].str.len() == 1
        ).groupby(postc_seq.index).agg(integrity_strictness) | (
            postc_seq.str[-1] == "-"
        ).groupby(
            postc_seq.index
        ).agg(
            integrity_strictness
        )

        internal_cuts = self.df["sequence"].str.split(rf"{enzyme_pattern}")
        self.df.loc[:, "missed_cleavages"] = (
            internal_cuts.apply(len)
            - internal_cuts.apply(lambda row: "" in row).astype(int)
            - 1
        )

    def calc_masses_offsets_and_composition(self):
        """Calculate chemical composition theoretical masses, and mass-to-charge ratios.

        Offsets are calculated between theoretical and experimental mass-to-charge ratio.
        Operations are performed inplace on self.df
        """
        all_compositions = {}
        for aa in self.IUPAC_AAS:
            self.cc.use(sequence=aa)
            all_compositions[aa] = self.cc.copy()
        for mod in self.mod_dict.keys():
            all_compositions[mod] = self.mod_mapper.name_to_composition(mod)[0]

        compositions, mono_masses = get_compositions_and_monoisotopic_masses(
            sequences=self.df["sequence"].to_numpy(dtype=str),
            modifications=self.df["modifications"].to_numpy(dtype=str),
            compositions=all_compositions,
            isotopic_distributions=self.cc.isotopic_distributions,
        )
        self.df.loc[:, "chemical_composition"] = compositions
        self.df.loc[:, "ucalc_mass"] = mono_masses

        with mp.Pool(
            self.params.get("cpus", mp.cpu_count() - 1),
            initializer=init_custom_cc,
            initargs=(
                get_isotopologue_accuracy,
                self.PROTON,
            ),
        ) as pool:
            acc = pool.starmap(
                get_isotopologue_accuracy,
                zip(
                    self.df["chemical_composition"].values,
                    self.df["charge"].astype(int).values,
                    self.df["exp_mz"].values,
                ),
            )
        self.df.loc[:, "accuracy_ppm"] = acc
        self.df.loc[:, "ucalc_mz"] = self._calc_mz(
            mass=self.df["ucalc_mass"], charge=self.df["charge"]
        )
        self.df.loc[:, "accuracy_ppm_C12"] = (
            (self.df["exp_mz"].astype(float) - self.df["ucalc_mz"])
            / self.df["ucalc_mz"]
            * 1e6
        )
        # Clear rows with non-mappable mods
        for mod in self.non_mappable_mods:
            self.df.loc[
                self.df["modifications"].str.contains(mod),
                (
                    "chemical_composition",
                    "ucalc_mass",
                    "ucalc_mz",
                    "accuracy_ppm",
                    "accuracy_ppm_C12",
                ),
            ] = np.nan
        self.df.loc[:, "exp_mass"] = self._calc_mass(
            mz=self.df["exp_mz"], charge=self.df["charge"]
        )
        self.df.loc[:, "mass_delta"] = self.df["exp_mass"] - self.df["ucalc_mass"]

    def get_meta_info(self):
        """Extract meta information.

        Experimental mass-to-charge ratios, retention times, file names,
        and spectrum titles are added.
        Operations are performed inplace on self.df
        """
        rt_lookup = self._read_meta_info_lookup_file()
        self.df["spectrum_id"] = self.df["spectrum_id"].astype(int)
        if self.style in ("comet_style_1", "omssa_style_1"):
            logger.warning(
                "This engine does not provide retention time information. Grouping only by Spectrum ID. This may cause problems when working with multi-file inputs."
            )
            for name, grp in self.df.groupby("spectrum_id"):
                mappable_within_precision = list(rt_lookup[name].keys())
                if len(mappable_within_precision) == 1:
                    self.df.loc[
                        grp.index,
                        ("raw_data_location", "exp_mz", "retention_time_seconds"),
                    ] = rt_lookup[name][mappable_within_precision[0]] + [
                        mappable_within_precision[0]
                    ]
                else:
                    logger.error(
                        f"Could not uniquely assign meta data to spectrum id {name}."
                    )
        else:
            self.df["retention_time_seconds"] = self.df[
                "retention_time_seconds"
            ].astype(float)
            for name, grp in self.df.groupby(["spectrum_id", "retention_time_seconds"]):
                meta_rts = rt_lookup[name[0]].keys()
                mappable_within_precision = [
                    rt
                    for rt in meta_rts
                    if abs(name[1] - rt) <= 10**-self.rt_truncate_precision
                ]
                if len(mappable_within_precision) == 1:
                    self.df.loc[
                        grp.index,
                        ("raw_data_location", "exp_mz", "retention_time_seconds"),
                    ] = rt_lookup[name[0]][mappable_within_precision[0]] + [
                        mappable_within_precision[0]
                    ]
                else:
                    logger.error(
                        f"Could not uniquely assign meta data to spectrum id, retention time {name}."
                    )

        self.df.loc[:, "spectrum_title"] = (
            self.df["raw_data_location"]
            + "."
            + self.df["spectrum_id"].astype(str)
            + "."
            + self.df["spectrum_id"].astype(str)
            + "."
            + self.df["charge"].astype(str)
        )

    def add_ranks(self):
        """Ranks are calculated based on the engine scoring column at Spectrum ID level.

        Operations are performed inplace on self.df
        """
        eng_name = self.df["search_engine"].unique()[0]
        score_col = self.params["validation_score_field"][eng_name]
        top_is_highest = self.params["bigger_scores_better"][eng_name]
        ranking_needs_to_be_ascending = False if top_is_highest is True else True
        self.df.loc[:, score_col] = self.df[score_col].astype(float)
        self.df.loc[:, "rank"] = (
            self.df.groupby("spectrum_id")[score_col]
            .rank(ascending=ranking_needs_to_be_ascending, method="min")
            .astype(int)
        )

    def add_decoy_identity(self):
        """Add boolean decoy state if designated decoy prefix is in Protein IDs.

        Also marks peptides which were assigned decoy but were immutable during
        target decoy generation.
        Operations are performed inplace on self.df
        """
        decoy_tag = self.params.get("decoy_tag", "decoy_")
        self.df.loc[:, "is_decoy"] = self.df["protein_id"].str.contains(decoy_tag)
        if self.immutable_peptides is not None:
            auto = ahocorasick.Automaton()
            for seq in self.immutable_peptides:
                auto.add_word(seq, seq)
            auto.make_automaton()
            self.df.loc[:, "is_immutable"] = [
                sum([len(match) for _, match in auto.iter_long(seq)]) == len(seq)
                for seq in self.df["sequence"]
            ]
        else:
            self.df.loc[:, "is_immutable"] = False

    def process_unify_style(self):
        """Combine all additional operations that are needed to calculate new columns and sanitize the dataframe.

        Operations are performed inplace on self.df
        """
        self.df.drop_duplicates(inplace=True, ignore_index=True)
        self.clean_up_modifications()
        self.assert_only_iupac_and_missing_aas()
        self.add_protein_ids()
        self.get_meta_info()
        self.calc_masses_offsets_and_composition()
        self.check_enzyme_specificity()
        self.add_ranks()
        self.add_decoy_identity()
        self.sanitize()



"""Quant parser."""
import re
from pathlib import Path

import pandas as pd
from chemical_composition import ChemicalComposition

from pyiohat.parsers.quant_base_parser import QuantBaseParser
from pyiohat.parsers.misc import get_compositions_and_monoisotopic_masses


class FlashLFQ_1_2_0_Parser(QuantBaseParser):
    """File parser for Flash LFQ."""

    def __init__(self, *args, **kwargs):
        """Initialize parser.

        Reads in data file and provides mappings.
        """
        super().__init__(*args, **kwargs)
        self.style = "flash_lfq_style_1"
        self.mapping_dict = {
            v: k
            for k, v in self.param_mapper.get_default_params(style=self.style)[
                "header_translations"
            ]["translated_value"].items()
        }
        self.df = pd.read_csv(self.input_file, delimiter="\t")
        self.df.rename(columns=self.mapping_dict, inplace=True)
        self.round_precision = 5

        self.rt_to_spec_id = {}
        self.filestem_to_path = {}
        for key, val in self.rt_lookup.items():
            for key2, val2 in val.items():
                self.rt_to_spec_id.setdefault(val2[0], {})[
                    round(key2, self.round_precision)
                ] = key
                if val2[0] not in self.filestem_to_path:
                    self.filestem_to_path[Path(val2[0]).stem] = val2[0]

            # round(list(d.keys())[0], self.round_precision): key
            # for key, d in self.rt_lookup.items()
        # }
        self.cc = ChemicalComposition()
        self.IUPAC_AAS = tuple("ACDEFGHIKLMNPQRSTUVWY")

    @classmethod
    def check_parser_compatibility(cls, file):
        """Assert compatibility between file and parser.

        Args:
            file (str): path to input file

        Returns:
            bool: True if parser and file are compatible

        """
        is_tsv = file.as_posix().endswith(".tsv")
        flash_lfq_columns = {
            "File Name",
            "Base Sequence",
            "Full Sequence",
            "Protein Group",
            "Peptide Monoisotopic Mass",
            "MS2 Retention Time",
            "Precursor Charge",
            "Theoretical MZ",
            "Peak intensity",
            "Peak RT Start",
            "Peak RT Apex",
            "Peak RT End",
            "Peak MZ",
            "Peak Charge",
            "Num Charge States Observed",
            "Peak Detection Type",
            "MBR Score",
            "PSMs Mapped",
            "Base Sequences Mapped",
            "Full Sequences Mapped",
            "Peak Split Valley RT",
            "Peak Apex Mass Error (ppm)",
        }
        with open(file.as_posix()) as f:
            head = set(f.readline().replace("\n", "").split("\t"))
        headers_match = len(flash_lfq_columns.difference(head)) == 0
        return is_tsv and headers_match

    def unify(self):
        """Primary method to read and unify engine output.

        Returns:
            self.df (pd.DataFrame): unified dataframe
        """
        # TODO: fill this
        # raise NotImplementedError
        # do column conversion here
        self.df["spectrum_id"] = -1
        self.df["ident_reference"] = -1
        self.df["retention_time_seconds"] = -1
        self.df["accuracy_ppm_C12"] = -1
        self.df["quant_score"] = -1
        self.df["fwhm"] = -1
        self.df["label"] = "LabelFree"
        self.df["quant_group"] = ""
        self.df["ccs"] = -1
        self.df.loc[self.df["flashlfq:mbr_score"].isna(), "flashlfq:mbr_score"] = -1

        self.get_chemical_composition()
        self.get_meta_info()
        self.calculate_accuracy()
        self.df["quant_run_id"] = "FlashLFQ"
        self.df["quant_engines"] = "FlashLFQ_1_2_0"

        self.process_unify_style()
        return self.df

    def get_meta_info(self):
        self.df["raw_filename"] = self.df["raw_filename"].map(self.filestem_to_path)
        rounded_rts = (self.df["flashlfq:ms2_retention_time"] / 60).apply(
            round, args=(self.round_precision,)
        )
        rounded_rts = pd.DataFrame({"file": self.df["raw_filename"], "rt": rounded_rts})

        self.df["ident_reference"] = [
            self.rt_to_spec_id[f][r] for i, (f, r) in rounded_rts.iterrows()
        ]

    def get_chemical_composition(self):
        mods = self.df["flashlfq:full_sequence"].apply(self.translate_mods)
        seqs = self.df["trivial_name"]

        # move to base parser?
        all_compositions = {}
        for aa in self.IUPAC_AAS:
            self.cc.use(sequence=aa)
            all_compositions[aa] = self.cc.copy()
        for mod in self.mod_dict.keys():
            all_compositions[mod] = self.mod_mapper.name_to_composition(mod)[0]

        compositions, mono_masses = get_compositions_and_monoisotopic_masses(
            sequences=seqs.to_numpy(dtype=str),
            modifications=mods.to_numpy(dtype=str),
            compositions=all_compositions,
            isotopic_distributions=self.cc.isotopic_distributions,
        )
        self.df.loc[:, "chemical_composition"] = compositions

    def translate_mods(self, full_sequence):
        """Extract modifications from full_sequence and format as {mod_1}:{pos_1};{mod_n}:{pos_n}.

        Args:
            full_sequence (str): sequence including mod (e.g. ELVISC[Carbamidomethyl]M[Oxidation])

        Returns:
            str: extracted mods as described in summary
        """
        # TODO extract C-terminal mods, position should be seq_len + 1
        cumulative_match_length = 0
        regex = re.compile(r"\[(.*?)\]")
        mods = []
        for match in regex.finditer(full_sequence):
            mods.append(f"{match.group(1)}:{match.start() - cumulative_match_length}")
            cumulative_match_length += len(match.group())
        return ";".join(mods)



#!/usr/bin/env python
import pytest

from pyiohat.parsers.ident.msgfplus_2021_03_22_parser import (
    MSGFPlus_2021_03_22_Parser,
)


def test_engine_parsers_msgfplus_init():
    input_file = pytest._test_path / "data" / "BSA1_msgfplus_2021_03_22.mzid"
    parser = MSGFPlus_2021_03_22_Parser(
        input_file,
        params={
            "cpus": 2,
            "enzyme": "(?<=[KR])(?![P])",
            "terminal_cleavage_site_integrity": "any",
            "validation_score_field": {"msgfplus_2021_03_22": "ms-gf:spec_evalue"},
            "bigger_scores_better": {"msgfplus_2021_03_22": False},
            "modifications": [
                {
                    "aa": "M",
                    "type": "opt",
                    "position": "any",
                    "name": "Oxidation",
                },
                {
                    "aa": "C",
                    "type": "fix",
                    "position": "any",
                    "name": "Carbamidomethyl",
                },
                {
                    "aa": "*",
                    "type": "opt",
                    "position": "Prot-N-term",
                    "name": "Acetyl",
                },
            ],
        },
    )


def test_engine_parsers_msgfplus_check_parser_compatibility():
    msgf_parser_class = MSGFPlus_2021_03_22_Parser
    input_file = pytest._test_path / "data" / "BSA1_msgfplus_2021_03_22.mzid"
    assert msgf_parser_class.check_parser_compatibility(input_file) is True


def test_engine_parsers_msgfplus_check_parser_compatibility_fail_with_omssa_file():
    msgf_parser_class = MSGFPlus_2021_03_22_Parser
    input_file = (
        pytest._test_path / "data" / "test_Creinhardtii_QE_pH11_omssa_2_1_9.csv"
    )
    assert msgf_parser_class.check_parser_compatibility(input_file) is False


def test_engine_parsers_msgfplus_check_dataframe_integrity():
    input_file = pytest._test_path / "data" / "BSA1_msgfplus_2021_03_22.mzid"
    rt_lookup_path = pytest._test_path / "data" / "BSA1_ursgal_lookup.csv"
    db_path = pytest._test_path / "data" / "BSA.fasta"

    parser = MSGFPlus_2021_03_22_Parser(
        input_file,
        params={
            "cpus": 2,
            "rt_pickle_name": rt_lookup_path,
            "database": db_path,
            "enzyme": "(?<=[KR])(?![P])",
            "terminal_cleavage_site_integrity": "any",
            "validation_score_field": {"msgfplus_2021_03_22": "ms-gf:spec_evalue"},
            "bigger_scores_better": {"msgfplus_2021_03_22": False},
            "modifications": [
                {
                    "aa": "M",
                    "type": "opt",
                    "position": "any",
                    "name": "Oxidation",
                },
                {
                    "aa": "C",
                    "type": "fix",
                    "position": "any",
                    "name": "Carbamidomethyl",
                },
                {
                    "aa": "*",
                    "type": "opt",
                    "position": "Prot-N-term",
                    "name": "Acetyl",
                },
            ],
        },
    )
    df = parser.unify()
    print(df)
    assert pytest.approx(df["exp_mz"].mean()) == 488.0319
    assert len(df) == 92
    assert pytest.approx(df["ucalc_mz"].mean()) == 488.03167
    assert (df["raw_data_location"] == "path/for/glory.mzML").all()
    assert df["modifications"].str.contains("Acetyl:0").sum() == 0
    assert df["modifications"].str.contains("Oxidation:").sum() == 0
    assert (
        df["modifications"].str.count("Carbamidomethyl:")
        == df["sequence"].str.count("C")
    ).all()
    assert df["modifications"].str.count(":").sum() == 71


def test_engine_parsers_msgfplus_check_dataframe_integrity_unknown_mod():
    input_file = (
        pytest._test_path / "data" / "BSA1_msgfplus_2021_03_22_unknown_mod.mzid"
    )
    rt_lookup_path = pytest._test_path / "data" / "BSA1_ursgal_lookup.csv"
    db_path = pytest._test_path / "data" / "BSA.fasta"

    parser = MSGFPlus_2021_03_22_Parser(
        input_file,
        params={
            "cpus": 2,
            "rt_pickle_name": rt_lookup_path,
            "database": db_path,
            "enzyme": "(?<=[KR])(?![P])",
            "terminal_cleavage_site_integrity": "any",
            "validation_score_field": {"msgfplus_2021_03_22": "ms-gf:spec_evalue"},
            "bigger_scores_better": {"msgfplus_2021_03_22": False},
            "modifications": [
                {
                    "aa": "M",
                    "type": "opt",
                    "position": "any",
                    "name": "Oxidation",
                },
                {
                    "aa": "C",
                    "type": "opt",
                    "position": "any",
                    "name": "Carbamidomethyl",
                },
                {
                    "aa": "C",
                    "type": "opt",
                    "position": "any",
                    "name": "DTB-IAA",
                },
                {
                    "aa": "*",
                    "type": "opt",
                    "position": "Prot-N-term",
                    "name": "Acetyl",
                },
            ],
        },
    )
    df = parser.unify()

    assert len(df) == 92
    # Currently this excludes two peptides and is reduced
    assert pytest.approx(df["ucalc_mz"].mean()) == 486.56
    assert (df["raw_data_location"] == "path/for/glory.mzML").all()
    assert df["modifications"].str.contains("Acetyl:0").sum() == 0
    assert df["modifications"].str.contains("Oxidation:").sum() == 0
    # the unknown modification is actually DTB-IAA
    assert df["modifications"].str.contains("DTB-IAA:3").sum() == 2
    assert (
        df[df["sequence"] != "EACFAVEGPK"]["modifications"].str.count(
            "Carbamidomethyl:"
        )
        == df[df["sequence"] != "EACFAVEGPK"]["sequence"].str.count("C")
    ).all()
    assert df["modifications"].str.count(":").sum() == 71


def test_engine_parsers_msgfplus_get_version():
    input_file = pytest._test_path / "data" / "BSA1_msgfplus_2021_03_22.mzid"
    parser = MSGFPlus_2021_03_22_Parser(input_file=input_file, params=None)
    version = parser.get_version()

    assert version == "msgfplus_2021_03_22"


def test_engine_parsers_msgfplus_get_peptide_lookup():
    input_file = pytest._test_path / "data" / "BSA1_msgfplus_2021_03_22.mzid"
    parser = MSGFPlus_2021_03_22_Parser(input_file=input_file, params=None)
    peptide_lookup = parser.get_peptide_lookup()

    assert len(peptide_lookup) == 24
    for pep in peptide_lookup:
        assert len(peptide_lookup[pep]) == 2
        assert pep.startswith("Pep_")
    assert peptide_lookup["Pep_LVTDLTK"] == {"sequence": "LVTDLTK", "modifications": ""}
    assert peptide_lookup["Pep_LVVSTQTALA"] == {
        "sequence": "LVVSTQTALA",
        "modifications": "",
    }


def test_engine_parsers_msgfplus_get_spec_records():
    input_file = pytest._test_path / "data" / "BSA1_msgfplus_2021_03_22.mzid"
    parser = MSGFPlus_2021_03_22_Parser(input_file=input_file, params=None)
    parser.peptide_lookup = parser.get_peptide_lookup()
    spec_records = parser.get_spec_records()

    assert len(spec_records) == 92
    assert spec_records[0] == {
        "ms-gf:raw_score": "40",
        "ms-gf:denovoscore": "40",
        "ms-gf:spec_evalue": "4.4458354E-15",
        "ms-gf:evalue": "2.6986221E-12",
        "ms-gf:num_matched_ions": "3",
        "charge": "2",
        "exp_mz": "722.3272094726562",
        "calc_mz": "722.3246459960938",
        "sequence": "YICDNQDTISSK",
        "modifications": "Carbamidomethyl:3",
        "spectrum_title": "glory.2791.2791.2",
        "spectrum_id": "2791",
        "retention_time_seconds": "1918.6086",
    }
    assert spec_records[44] == {
        "ms-gf:raw_score": "26",
        "ms-gf:denovoscore": "32",
        "ms-gf:spec_evalue": "2.2523169E-8",
        "ms-gf:evalue": "1.36715635E-5",
        "ms-gf:num_matched_ions": "1",
        "charge": "3",
        "exp_mz": "325.4911804199219",
        "calc_mz": "325.4907531738281",
        "sequence": "DLGEEHFK",
        "modifications": "",
        "spectrum_title": "glory.mzML.2663.2663.3",
        "spectrum_id": "2663",
        "retention_time_seconds": "1840.7933333333335",
    }
